import rich.markup
import rich_click as click
from rich.console import Console

from union._config import _DEFAULT_DOMAIN, _DEFAULT_PROJECT_BYOC


@click.group()
def start():
    """Start a resource."""


@start.command()
@click.option("--name", type=str, required=True)
@click.option("--project", default=_DEFAULT_PROJECT_BYOC, help="Project name")
@click.option("--domain", default=_DEFAULT_DOMAIN, help="Domain name")
def workspace(name: str, project: str, domain: str):
    from union._async import run_sync
    from union.configuration import UnionAIPlugin
    from union.workspace._vscode_remote import WorkspaceRemote

    remote = UnionAIPlugin.get_remote(config=None, project=project, domain=domain)
    ws_remote = WorkspaceRemote(default_project=project, default_domain=domain, union_remote=remote)

    console = Console()
    console.print(f"Starting workspace '{name}'")
    with console.status("Provisioning workspace", spinner="dots"):
        try:
            response = run_sync(ws_remote.start_workspace_instance, workspace_definition_name=name)
        except Exception as exc:
            error_msg = exc.details() if hasattr(exc, "details") else str(exc)
            console.print(f"Error provisioning workspace: {rich.markup.escape(error_msg)}", style="red")
            raise exc

    url = f"https://{remote.config.platform.endpoint}/{response.workspace_instance.spec.uri}"
    loading_url = f"https://{remote.config.platform.endpoint}/loading?type=workspace&url={url}"
    console.print(f"\n🚀 Workspace started: [link={loading_url}]Open VSCode in Browser[/link]")



from union.internal.common import identifier_pb2 as _identifier_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class OverallStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNSPECIFIED: _ClassVar[OverallStatus]
    PARTIALLY_PRESENT: _ClassVar[OverallStatus]
    FULLY_PRESENT: _ClassVar[OverallStatus]
    UNKNOWN_STATUS: _ClassVar[OverallStatus]

class SecretPresenceStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN: _ClassVar[SecretPresenceStatus]
    MISSING: _ClassVar[SecretPresenceStatus]
    PRESENT: _ClassVar[SecretPresenceStatus]
UNSPECIFIED: OverallStatus
PARTIALLY_PRESENT: OverallStatus
FULLY_PRESENT: OverallStatus
UNKNOWN_STATUS: OverallStatus
UNKNOWN: SecretPresenceStatus
MISSING: SecretPresenceStatus
PRESENT: SecretPresenceStatus

class SecretSpec(_message.Message):
    __slots__ = ["string_value", "binary_value"]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    BINARY_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    binary_value: bytes
    def __init__(self, string_value: _Optional[str] = ..., binary_value: _Optional[bytes] = ...) -> None: ...

class SecretIdentifier(_message.Message):
    __slots__ = ["name", "organization", "domain", "project"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    name: str
    organization: str
    domain: str
    project: str
    def __init__(self, name: _Optional[str] = ..., organization: _Optional[str] = ..., domain: _Optional[str] = ..., project: _Optional[str] = ...) -> None: ...

class SecretMetadata(_message.Message):
    __slots__ = ["created_time", "secret_status"]
    CREATED_TIME_FIELD_NUMBER: _ClassVar[int]
    SECRET_STATUS_FIELD_NUMBER: _ClassVar[int]
    created_time: _timestamp_pb2.Timestamp
    secret_status: SecretStatus
    def __init__(self, created_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., secret_status: _Optional[_Union[SecretStatus, _Mapping]] = ...) -> None: ...

class SecretStatus(_message.Message):
    __slots__ = ["overall_status", "cluster_status"]
    OVERALL_STATUS_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_STATUS_FIELD_NUMBER: _ClassVar[int]
    overall_status: OverallStatus
    cluster_status: _containers.RepeatedCompositeFieldContainer[ClusterSecretStatus]
    def __init__(self, overall_status: _Optional[_Union[OverallStatus, str]] = ..., cluster_status: _Optional[_Iterable[_Union[ClusterSecretStatus, _Mapping]]] = ...) -> None: ...

class ClusterSecretStatus(_message.Message):
    __slots__ = ["cluster", "presence_status"]
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    PRESENCE_STATUS_FIELD_NUMBER: _ClassVar[int]
    cluster: _identifier_pb2.ClusterIdentifier
    presence_status: SecretPresenceStatus
    def __init__(self, cluster: _Optional[_Union[_identifier_pb2.ClusterIdentifier, _Mapping]] = ..., presence_status: _Optional[_Union[SecretPresenceStatus, str]] = ...) -> None: ...

class Secret(_message.Message):
    __slots__ = ["id", "secret_metadata"]
    ID_FIELD_NUMBER: _ClassVar[int]
    SECRET_METADATA_FIELD_NUMBER: _ClassVar[int]
    id: SecretIdentifier
    secret_metadata: SecretMetadata
    def __init__(self, id: _Optional[_Union[SecretIdentifier, _Mapping]] = ..., secret_metadata: _Optional[_Union[SecretMetadata, _Mapping]] = ...) -> None: ...



import flytekit.core.artifact  # noqa: F401 place this line first

from union.artifacts._artifact import Artifact
from union.artifacts._card import DataCard, ModelCard
from union.artifacts._triggers import OnArtifact

__all__ = ["Artifact", "OnArtifact", "DataCard", "ModelCard"]



from union.internal.app import app_definition_pb2 as _app_definition_pb2
from union.internal.app import replica_definition_pb2 as _replica_definition_pb2
from union.internal.logs.dataplane import payload_pb2 as _payload_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class TailLogsRequest(_message.Message):
    __slots__ = ["app_id", "replica_id"]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    REPLICA_ID_FIELD_NUMBER: _ClassVar[int]
    app_id: _app_definition_pb2.Identifier
    replica_id: _replica_definition_pb2.ReplicaIdentifier
    def __init__(self, app_id: _Optional[_Union[_app_definition_pb2.Identifier, _Mapping]] = ..., replica_id: _Optional[_Union[_replica_definition_pb2.ReplicaIdentifier, _Mapping]] = ...) -> None: ...

class ReplicaIdentifierList(_message.Message):
    __slots__ = ["replicas"]
    REPLICAS_FIELD_NUMBER: _ClassVar[int]
    replicas: _containers.RepeatedCompositeFieldContainer[_replica_definition_pb2.ReplicaIdentifier]
    def __init__(self, replicas: _Optional[_Iterable[_Union[_replica_definition_pb2.ReplicaIdentifier, _Mapping]]] = ...) -> None: ...

class LogLines(_message.Message):
    __slots__ = ["lines", "replica_id", "structured_lines"]
    LINES_FIELD_NUMBER: _ClassVar[int]
    REPLICA_ID_FIELD_NUMBER: _ClassVar[int]
    STRUCTURED_LINES_FIELD_NUMBER: _ClassVar[int]
    lines: _containers.RepeatedScalarFieldContainer[str]
    replica_id: _replica_definition_pb2.ReplicaIdentifier
    structured_lines: _containers.RepeatedCompositeFieldContainer[_payload_pb2.LogLine]
    def __init__(self, lines: _Optional[_Iterable[str]] = ..., replica_id: _Optional[_Union[_replica_definition_pb2.ReplicaIdentifier, _Mapping]] = ..., structured_lines: _Optional[_Iterable[_Union[_payload_pb2.LogLine, _Mapping]]] = ...) -> None: ...

class LogLinesBatch(_message.Message):
    __slots__ = ["logs"]
    LOGS_FIELD_NUMBER: _ClassVar[int]
    logs: _containers.RepeatedCompositeFieldContainer[LogLines]
    def __init__(self, logs: _Optional[_Iterable[_Union[LogLines, _Mapping]]] = ...) -> None: ...

class TailLogsResponse(_message.Message):
    __slots__ = ["replicas", "log_lines", "batches"]
    REPLICAS_FIELD_NUMBER: _ClassVar[int]
    LOG_LINES_FIELD_NUMBER: _ClassVar[int]
    BATCHES_FIELD_NUMBER: _ClassVar[int]
    replicas: ReplicaIdentifierList
    log_lines: _payload_pb2.LogLines
    batches: LogLinesBatch
    def __init__(self, replicas: _Optional[_Union[ReplicaIdentifierList, _Mapping]] = ..., log_lines: _Optional[_Union[_payload_pb2.LogLines, _Mapping]] = ..., batches: _Optional[_Union[LogLinesBatch, _Mapping]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: identity/user_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from union.internal.identity import user_payload_pb2 as identity_dot_user__payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bidentity/user_service.proto\x12\x11\x63loudidl.identity\x1a\x1cgoogle/api/annotations.proto\x1a\x1bidentity/user_payload.proto2\x9c\x06\n\x0bUserService\x12s\n\nCreateUser\x12$.cloudidl.identity.CreateUserRequest\x1a%.cloudidl.identity.CreateUserResponse\"\x18\x82\xd3\xe4\x93\x02\x12\x1a\r/users/api/v1:\x01*\x12t\n\x07GetUser\x12!.cloudidl.identity.GetUserRequest\x1a\".cloudidl.identity.GetUserResponse\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/users/api/v1/{id.subject}\x12\x80\x01\n\nDeleteUser\x12$.cloudidl.identity.DeleteUserRequest\x1a%.cloudidl.identity.DeleteUserResponse\"%\x82\xd3\xe4\x93\x02\x1f*\x1a/users/api/v1/{id.subject}:\x01*\x12m\n\tListUsers\x12#.cloudidl.identity.ListUsersRequest\x1a$.cloudidl.identity.ListUsersResponse\"\x15\x82\xd3\xe4\x93\x02\x0f\x12\r/users/api/v1\x12\x82\x01\n\x0eListUsersCount\x12(.cloudidl.identity.ListUsersCountRequest\x1a).cloudidl.identity.ListUsersCountResponse\"\x1b\x82\xd3\xe4\x93\x02\x15\x12\x13/users/api/v1/count\x12\xaa\x01\n\x16GetUserAndGroupsForOrg\x12\x30.cloudidl.identity.GetUserAndGroupsForOrgRequest\x1a\x31.cloudidl.identity.GetUserAndGroupsForOrgResponse\"+\x82\xd3\xe4\x93\x02%\x12#/users/api/v1/{organization}/groupsB\xbd\x01\n\x15\x63om.cloudidl.identityB\x10UserServiceProtoH\x02P\x01Z+github.com/unionai/cloud/gen/pb-go/identity\xa2\x02\x03\x43IX\xaa\x02\x11\x43loudidl.Identity\xca\x02\x11\x43loudidl\\Identity\xe2\x02\x1d\x43loudidl\\Identity\\GPBMetadata\xea\x02\x12\x43loudidl::Identityb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'identity.user_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\025com.cloudidl.identityB\020UserServiceProtoH\002P\001Z+github.com/unionai/cloud/gen/pb-go/identity\242\002\003CIX\252\002\021Cloudidl.Identity\312\002\021Cloudidl\\Identity\342\002\035Cloudidl\\Identity\\GPBMetadata\352\002\022Cloudidl::Identity'
  _USERSERVICE.methods_by_name['CreateUser']._options = None
  _USERSERVICE.methods_by_name['CreateUser']._serialized_options = b'\202\323\344\223\002\022\032\r/users/api/v1:\001*'
  _USERSERVICE.methods_by_name['GetUser']._options = None
  _USERSERVICE.methods_by_name['GetUser']._serialized_options = b'\202\323\344\223\002\034\022\032/users/api/v1/{id.subject}'
  _USERSERVICE.methods_by_name['DeleteUser']._options = None
  _USERSERVICE.methods_by_name['DeleteUser']._serialized_options = b'\202\323\344\223\002\037*\032/users/api/v1/{id.subject}:\001*'
  _USERSERVICE.methods_by_name['ListUsers']._options = None
  _USERSERVICE.methods_by_name['ListUsers']._serialized_options = b'\202\323\344\223\002\017\022\r/users/api/v1'
  _USERSERVICE.methods_by_name['ListUsersCount']._options = None
  _USERSERVICE.methods_by_name['ListUsersCount']._serialized_options = b'\202\323\344\223\002\025\022\023/users/api/v1/count'
  _USERSERVICE.methods_by_name['GetUserAndGroupsForOrg']._options = None
  _USERSERVICE.methods_by_name['GetUserAndGroupsForOrg']._serialized_options = b'\202\323\344\223\002%\022#/users/api/v1/{organization}/groups'
  _globals['_USERSERVICE']._serialized_start=110
  _globals['_USERSERVICE']._serialized_end=906
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class LogLineOriginator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN: _ClassVar[LogLineOriginator]
    USER: _ClassVar[LogLineOriginator]
    SYSTEM: _ClassVar[LogLineOriginator]

class LogsSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    LIVE_OR_PERSISTED: _ClassVar[LogsSource]
    LIVE_ONLY: _ClassVar[LogsSource]
    PERSISTED_ONLY: _ClassVar[LogsSource]
UNKNOWN: LogLineOriginator
USER: LogLineOriginator
SYSTEM: LogLineOriginator
LIVE_OR_PERSISTED: LogsSource
LIVE_ONLY: LogsSource
PERSISTED_ONLY: LogsSource

class PodResource(_message.Message):
    __slots__ = ["namespace", "name", "container"]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    name: str
    container: str
    def __init__(self, namespace: _Optional[str] = ..., name: _Optional[str] = ..., container: _Optional[str] = ...) -> None: ...

class TailTaskExecutionLogsRequest(_message.Message):
    __slots__ = ["logging_context", "source", "no_follow"]
    LOGGING_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    NO_FOLLOW_FIELD_NUMBER: _ClassVar[int]
    logging_context: LoggingContext
    source: LogsSource
    no_follow: bool
    def __init__(self, logging_context: _Optional[_Union[LoggingContext, _Mapping]] = ..., source: _Optional[_Union[LogsSource, str]] = ..., no_follow: bool = ...) -> None: ...

class TailTaskExecutionLogsResponse(_message.Message):
    __slots__ = ["logs"]
    class Logs(_message.Message):
        __slots__ = ["lines", "structured_lines"]
        LINES_FIELD_NUMBER: _ClassVar[int]
        STRUCTURED_LINES_FIELD_NUMBER: _ClassVar[int]
        lines: _containers.RepeatedScalarFieldContainer[str]
        structured_lines: _containers.RepeatedCompositeFieldContainer[LogLine]
        def __init__(self, lines: _Optional[_Iterable[str]] = ..., structured_lines: _Optional[_Iterable[_Union[LogLine, _Mapping]]] = ...) -> None: ...
    LOGS_FIELD_NUMBER: _ClassVar[int]
    logs: TailTaskExecutionLogsResponse.Logs
    def __init__(self, logs: _Optional[_Union[TailTaskExecutionLogsResponse.Logs, _Mapping]] = ...) -> None: ...

class LoggingContext(_message.Message):
    __slots__ = ["cluster_name", "kubernetes_namespace", "kubernetes_pod_name", "kubernetes_container_name", "execution_attempt_start_time", "execution_attempt_end_time"]
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_POD_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_CONTAINER_NAME_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_ATTEMPT_START_TIME_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_ATTEMPT_END_TIME_FIELD_NUMBER: _ClassVar[int]
    cluster_name: str
    kubernetes_namespace: str
    kubernetes_pod_name: str
    kubernetes_container_name: str
    execution_attempt_start_time: _timestamp_pb2.Timestamp
    execution_attempt_end_time: _timestamp_pb2.Timestamp
    def __init__(self, cluster_name: _Optional[str] = ..., kubernetes_namespace: _Optional[str] = ..., kubernetes_pod_name: _Optional[str] = ..., kubernetes_container_name: _Optional[str] = ..., execution_attempt_start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., execution_attempt_end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ContainerIdentifier(_message.Message):
    __slots__ = ["cluster_name", "kubernetes_namespace", "kubernetes_pod_name", "kubernetes_container_name"]
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_POD_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_CONTAINER_NAME_FIELD_NUMBER: _ClassVar[int]
    cluster_name: str
    kubernetes_namespace: str
    kubernetes_pod_name: str
    kubernetes_container_name: str
    def __init__(self, cluster_name: _Optional[str] = ..., kubernetes_namespace: _Optional[str] = ..., kubernetes_pod_name: _Optional[str] = ..., kubernetes_container_name: _Optional[str] = ...) -> None: ...

class ContainerSelector(_message.Message):
    __slots__ = ["cluster_name", "kubernetes_namespace", "kubernetes_pod_name_prefix", "kubernetes_container_name", "kubernetes_pod_label_selector"]
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_POD_NAME_PREFIX_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_CONTAINER_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBERNETES_POD_LABEL_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    cluster_name: str
    kubernetes_namespace: str
    kubernetes_pod_name_prefix: str
    kubernetes_container_name: str
    kubernetes_pod_label_selector: str
    def __init__(self, cluster_name: _Optional[str] = ..., kubernetes_namespace: _Optional[str] = ..., kubernetes_pod_name_prefix: _Optional[str] = ..., kubernetes_container_name: _Optional[str] = ..., kubernetes_pod_label_selector: _Optional[str] = ...) -> None: ...

class LiveLogsOptions(_message.Message):
    __slots__ = ["log_pod_status", "log_timestamps"]
    LOG_POD_STATUS_FIELD_NUMBER: _ClassVar[int]
    LOG_TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
    log_pod_status: bool
    log_timestamps: bool
    def __init__(self, log_pod_status: bool = ..., log_timestamps: bool = ...) -> None: ...

class TailLogsRequest(_message.Message):
    __slots__ = ["container", "container_selector", "start_time", "end_time", "source", "live_logs_options"]
    CONTAINER_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_SELECTOR_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    LIVE_LOGS_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    container: ContainerIdentifier
    container_selector: ContainerSelector
    start_time: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    source: LogsSource
    live_logs_options: LiveLogsOptions
    def __init__(self, container: _Optional[_Union[ContainerIdentifier, _Mapping]] = ..., container_selector: _Optional[_Union[ContainerSelector, _Mapping]] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., source: _Optional[_Union[LogsSource, str]] = ..., live_logs_options: _Optional[_Union[LiveLogsOptions, _Mapping]] = ...) -> None: ...

class LogLine(_message.Message):
    __slots__ = ["timestamp", "message", "originator"]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    ORIGINATOR_FIELD_NUMBER: _ClassVar[int]
    timestamp: _timestamp_pb2.Timestamp
    message: str
    originator: LogLineOriginator
    def __init__(self, timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., message: _Optional[str] = ..., originator: _Optional[_Union[LogLineOriginator, str]] = ...) -> None: ...

class LogLines(_message.Message):
    __slots__ = ["lines", "container_index", "container", "structured_lines"]
    LINES_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_INDEX_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_FIELD_NUMBER: _ClassVar[int]
    STRUCTURED_LINES_FIELD_NUMBER: _ClassVar[int]
    lines: _containers.RepeatedScalarFieldContainer[str]
    container_index: int
    container: ContainerIdentifier
    structured_lines: _containers.RepeatedCompositeFieldContainer[LogLine]
    def __init__(self, lines: _Optional[_Iterable[str]] = ..., container_index: _Optional[int] = ..., container: _Optional[_Union[ContainerIdentifier, _Mapping]] = ..., structured_lines: _Optional[_Iterable[_Union[LogLine, _Mapping]]] = ...) -> None: ...

class LogContainersList(_message.Message):
    __slots__ = ["containers"]
    CONTAINERS_FIELD_NUMBER: _ClassVar[int]
    containers: _containers.RepeatedCompositeFieldContainer[ContainerIdentifier]
    def __init__(self, containers: _Optional[_Iterable[_Union[ContainerIdentifier, _Mapping]]] = ...) -> None: ...

class LogLinesBatch(_message.Message):
    __slots__ = ["logs"]
    LOGS_FIELD_NUMBER: _ClassVar[int]
    logs: _containers.RepeatedCompositeFieldContainer[LogLines]
    def __init__(self, logs: _Optional[_Iterable[_Union[LogLines, _Mapping]]] = ...) -> None: ...

class TailLogsResponse(_message.Message):
    __slots__ = ["containers", "log_lines_batch"]
    CONTAINERS_FIELD_NUMBER: _ClassVar[int]
    LOG_LINES_BATCH_FIELD_NUMBER: _ClassVar[int]
    containers: LogContainersList
    log_lines_batch: LogLinesBatch
    def __init__(self, containers: _Optional[_Union[LogContainersList, _Mapping]] = ..., log_lines_batch: _Optional[_Union[LogLinesBatch, _Mapping]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: secret/definition.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.common import identifier_pb2 as common_dot_identifier__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17secret/definition.proto\x12\x0f\x63loudidl.secret\x1a\x17\x63ommon/identifier.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x17validate/validate.proto\"d\n\nSecretSpec\x12#\n\x0cstring_value\x18\x01 \x01(\tH\x00R\x0bstringValue\x12#\n\x0c\x62inary_value\x18\x02 \x01(\x0cH\x00R\x0b\x62inaryValueB\x0c\n\x05value\x12\x03\xf8\x42\x01\"\x97\x01\n\x10SecretIdentifier\x12-\n\x04name\x18\x01 \x01(\tB\x19\xfa\x42\x16r\x14\x10\x01\x32\x10^[-a-zA-Z0-9_]+$R\x04name\x12\"\n\x0corganization\x18\x02 \x01(\tR\x0corganization\x12\x16\n\x06\x64omain\x18\x03 \x01(\tR\x06\x64omain\x12\x18\n\x07project\x18\x04 \x01(\tR\x07project\"\x93\x01\n\x0eSecretMetadata\x12=\n\x0c\x63reated_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0b\x63reatedTime\x12\x42\n\rsecret_status\x18\x02 \x01(\x0b\x32\x1d.cloudidl.secret.SecretStatusR\x0csecretStatus\"\xa2\x01\n\x0cSecretStatus\x12\x45\n\x0eoverall_status\x18\x01 \x01(\x0e\x32\x1e.cloudidl.secret.OverallStatusR\roverallStatus\x12K\n\x0e\x63luster_status\x18\x02 \x03(\x0b\x32$.cloudidl.secret.ClusterSecretStatusR\rclusterStatus\"\xa3\x01\n\x13\x43lusterSecretStatus\x12<\n\x07\x63luster\x18\x01 \x01(\x0b\x32\".cloudidl.common.ClusterIdentifierR\x07\x63luster\x12N\n\x0fpresence_status\x18\x02 \x01(\x0e\x32%.cloudidl.secret.SecretPresenceStatusR\x0epresenceStatus\"\x85\x01\n\x06Secret\x12\x31\n\x02id\x18\x01 \x01(\x0b\x32!.cloudidl.secret.SecretIdentifierR\x02id\x12H\n\x0fsecret_metadata\x18\x02 \x01(\x0b\x32\x1f.cloudidl.secret.SecretMetadataR\x0esecretMetadata*^\n\rOverallStatus\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x15\n\x11PARTIALLY_PRESENT\x10\x01\x12\x11\n\rFULLY_PRESENT\x10\x02\x12\x12\n\x0eUNKNOWN_STATUS\x10\x03*=\n\x14SecretPresenceStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07MISSING\x10\x01\x12\x0b\n\x07PRESENT\x10\x02\x42\xb0\x01\n\x13\x63om.cloudidl.secretB\x0f\x44\x65\x66initionProtoH\x02P\x01Z)github.com/unionai/cloud/gen/pb-go/secret\xa2\x02\x03\x43SX\xaa\x02\x0f\x43loudidl.Secret\xca\x02\x0f\x43loudidl\\Secret\xe2\x02\x1b\x43loudidl\\Secret\\GPBMetadata\xea\x02\x10\x43loudidl::Secretb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'secret.definition_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.cloudidl.secretB\017DefinitionProtoH\002P\001Z)github.com/unionai/cloud/gen/pb-go/secret\242\002\003CSX\252\002\017Cloudidl.Secret\312\002\017Cloudidl\\Secret\342\002\033Cloudidl\\Secret\\GPBMetadata\352\002\020Cloudidl::Secret'
  _SECRETSPEC.oneofs_by_name['value']._options = None
  _SECRETSPEC.oneofs_by_name['value']._serialized_options = b'\370B\001'
  _SECRETIDENTIFIER.fields_by_name['name']._options = None
  _SECRETIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\026r\024\020\0012\020^[-a-zA-Z0-9_]+$'
  _globals['_OVERALLSTATUS']._serialized_start=1000
  _globals['_OVERALLSTATUS']._serialized_end=1094
  _globals['_SECRETPRESENCESTATUS']._serialized_start=1096
  _globals['_SECRETPRESENCESTATUS']._serialized_end=1157
  _globals['_SECRETSPEC']._serialized_start=127
  _globals['_SECRETSPEC']._serialized_end=227
  _globals['_SECRETIDENTIFIER']._serialized_start=230
  _globals['_SECRETIDENTIFIER']._serialized_end=381
  _globals['_SECRETMETADATA']._serialized_start=384
  _globals['_SECRETMETADATA']._serialized_end=531
  _globals['_SECRETSTATUS']._serialized_start=534
  _globals['_SECRETSTATUS']._serialized_end=696
  _globals['_CLUSTERSECRETSTATUS']._serialized_start=699
  _globals['_CLUSTERSECRETSTATUS']._serialized_end=862
  _globals['_SECRET']._serialized_start=865
  _globals['_SECRET']._serialized_end=998
# @@protoc_insertion_point(module_scope)



import ast
import hashlib
import inspect
import textwrap

from flytekit import VersionParameters


class CacheFunctionBody:
    """
    A class that implements a versioning mechanism for functions by generating
    a SHA-256 hash of the function's source code combined with a salt.
    """

    def get_version(self, salt: str, params: VersionParameters) -> str:
        """
        This method generates a version string for a function by hashing the function's source code
        combined with a salt.

        :param salt: A string that is used to salt the hash.
        :param params: VersionParameters object that contains the parameters (e.g. function, ImageSpec, etc.) that are
                       used to generate the version.

        :return: A string that represents the version of the function.
        """
        source = inspect.getsource(params.func)
        dedented_source = textwrap.dedent(source)

        # Parse the source code into an Abstract Syntax Tree (AST)
        parsed_ast = ast.parse(dedented_source)

        # Convert the AST into a string representation
        ast_bytes = ast.dump(parsed_ast, include_attributes=False).encode("utf-8")

        # Combine the AST bytes with the salt (encoded into bytes)
        combined_data = ast_bytes + salt.encode("utf-8")

        # Return the SHA-256 hash of the combined data (AST + salt)
        return hashlib.sha256(combined_data).hexdigest()



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RuntimeMetadata(_message.Message):
    __slots__ = ["type", "version", "flavor"]
    class RuntimeType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        OTHER: _ClassVar[RuntimeMetadata.RuntimeType]
        FLYTE_SDK: _ClassVar[RuntimeMetadata.RuntimeType]
        UNION_SDK: _ClassVar[RuntimeMetadata.RuntimeType]
    OTHER: RuntimeMetadata.RuntimeType
    FLYTE_SDK: RuntimeMetadata.RuntimeType
    UNION_SDK: RuntimeMetadata.RuntimeType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    FLAVOR_FIELD_NUMBER: _ClassVar[int]
    type: RuntimeMetadata.RuntimeType
    version: str
    flavor: str
    def __init__(self, type: _Optional[_Union[RuntimeMetadata.RuntimeType, str]] = ..., version: _Optional[str] = ..., flavor: _Optional[str] = ...) -> None: ...



import typing
from typing import Tuple

import grpc
import rich_click as click
from flytekit.clients.auth_helper import get_authenticated_channel
from flytekit.configuration import PlatformConfig

from union._config import _get_organization
from union._interceptor import intercept_channel_with_org


def _get_channel_with_org(platform_config: PlatformConfig) -> Tuple[grpc.Channel, str]:
    """Construct authenticated channel that injects the org."""
    channel = get_authenticated_channel(platform_config)

    org = _get_organization(platform_config, channel)
    channel = intercept_channel_with_org(org, channel)
    return channel, org


def _validate_key_value_pairs(ctx, param, values) -> typing.Dict[str, str]:
    """
    Validate key value pairs, parses given values into a dictionary and validates that it is of type key=value.
    """
    if not values:
        return {}
    result = {}
    for value in values:
        if "=" not in value:
            raise click.BadParameter(f"Invalid key value pair: {value}")
        key, v = value.split("=", 1)
        result[key] = v
    return result


def _common_options() -> typing.List[click.Option]:
    return [
        click.Option(
            param_decls=["-p", "--project"],
            required=False,
            type=str,
            default="default",
            help="Project to operate on",
            show_default=True,
        ),
        click.Option(
            param_decls=["-d", "--domain"],
            required=False,
            type=str,
            default="development",
            help="Domain to operate on",
            show_default=True,
        ),
    ]


class InvokeBaseMixin:
    def invoke(self, ctx):
        try:
            return super().invoke(ctx)
        except grpc.RpcError as e:
            raise click.ClickException(f"Error invoking command: {e.details()}") from e
        except Exception as e:
            raise click.ClickException(f"Error invoking command: {e}") from e


class CommandBase(InvokeBaseMixin, click.RichCommand):
    def __init__(self, *args, **kwargs):
        if "params" not in kwargs:
            kwargs["params"] = []
        kwargs["params"].extend(_common_options())
        super().__init__(*args, **kwargs)


class GroupBase(InvokeBaseMixin, click.RichGroup):
    def __init__(self, *args, **kwargs):
        if "params" not in kwargs:
            kwargs["params"] = []
        kwargs["params"].extend(_common_options())
        super().__init__(*args, **kwargs)



"""Module to manage applications on Union."""

import asyncio
import logging
import os
import tarfile
import weakref
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, List, Optional

import grpc
from click import ClickException
from flyteidl.core.tasks_pb2 import Resources
from flytekit.configuration import Config
from flytekit.core.artifact import ArtifactQuery
from flytekit.models.core.types import BlobType
from rich.console import Console

from union._async import merge, run_sync
from union._config import _get_organization, _is_serverless_endpoint
from union.app import App
from union.app._models import AppSerializationSettings, Input, MaterializedInput, URLQuery
from union.cli._common import _get_channel_with_org
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.app.app_definition_pb2 import Condition, Identifier, Spec, Status
from union.internal.app.app_logs_payload_pb2 import TailLogsRequest, TailLogsResponse
from union.internal.app.app_logs_service_pb2_grpc import AppLogsServiceStub
from union.internal.app.app_payload_pb2 import (
    CreateRequest,
    CreateResponse,
    GetRequest,
    GetResponse,
    ListResponse,
    UpdateRequest,
    UpdateResponse,
    WatchRequest,
    WatchResponse,
)
from union.internal.app.app_payload_pb2 import (
    ListRequest as ListAppsRequest,
)
from union.internal.app.app_service_pb2_grpc import AppServiceStub
from union.internal.common.identifier_pb2 import ProjectIdentifier
from union.internal.common.list_pb2 import ListRequest
from union.remote._remote import UnionRemote

logger = logging.getLogger(__name__)

FILES_TAR_FILE_NAME = "include-files.tar.gz"


class AppRemote:
    def __init__(
        self,
        default_project: str,
        default_domain: str,
        config: Optional[Config] = None,
        union_remote: Optional[UnionRemote] = None,
    ):
        if union_remote is None:
            union_remote = UnionRemote(config=config, default_domain=default_domain, default_project=default_project)
            self._union_remote = union_remote
            self._union_remote_ref = weakref.ref(self._union_remote)
        else:
            # Union remote is passed in directly. We assume that the caller will have a reference to `AppRemote`.
            self._union_remote_ref = weakref.ref(union_remote)

        self.config = union_remote.config
        self.default_project = union_remote.default_project
        self.default_domain = union_remote.default_domain

    @property
    def union_remote(self) -> UnionRemote:
        union_remote = self._union_remote_ref()
        if union_remote is None:
            raise ValueError("Unable to find union remote")
        return union_remote

    def list(self) -> List[AppIDL]:
        def create_list_request(token: str):
            return ListAppsRequest(
                request=ListRequest(token=token),
                org=self.org,
                project=ProjectIdentifier(name=self.default_project, domain=self.default_domain, organization=self.org),
            )

        results = []
        response: ListResponse
        token, has_next = "", True

        while has_next:
            list_request = create_list_request(token=token)

            response = self.sync_client.List(list_request)
            token = response.token
            has_next = token != ""

            results.extend(response.apps)

        return results

    def deploy(self, app: App, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        try:
            orig_app_idl = self.get(name=app.name, project=project, domain=domain)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                return self.create(app, project=project, domain=domain)
            raise

        try:
            return self.update(app, project=project, domain=domain)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.ABORTED and "Either the change has already" in e.details():
                console = Console()
                console.print(f"✅ {app.name} app was already deployed with the current spec")

                _url = orig_app_idl.status.ingress.public_url
                url = f"[link={_url}]{_url}[/link]"

                console.print()
                console.print(f"🚀 Deployed Endpoint: {url}")

                return orig_app_idl
            raise

    def get(self, name: str, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        app_id = Identifier(
            org=self.org,
            project=project or self.default_project,
            domain=domain or self.default_domain,
            name=name,
        )
        get_app_request = GetRequest(app_id=app_id)

        get_response: GetResponse = self.sync_client.Get(get_app_request)
        app = get_response.app

        return app

    def start(self, name: str, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        app_idl = self.get(name, project=project, domain=domain)
        app_idl.spec.desired_state = Spec.DesiredState.DESIRED_STATE_STARTED

        def run_request():
            update_request = UpdateRequest(app=app_idl)
            update_response: UpdateResponse = self.sync_client.Update(update_request)
            return update_response.app

        initial_condition = None
        if app_idl.status.conditions:
            initial_condition = app_idl.status.conditions[-1]

        app_idl = self._watch_update(app_idl, run_request, initial_condition)
        return app_idl

    def update(self, app: App, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        additional_distribution = self.upload_files(app)
        materialized_input_values = self.materialize_values(app, project=project, domain=domain)
        settings = AppSerializationSettings(
            org=self.org,
            project=project or self.default_project,
            domain=domain or self.default_domain,
            additional_distribution=additional_distribution,
            desired_state=Spec.DesiredState.DESIRED_STATE_STARTED,
            materialized_inputs=materialized_input_values,
            is_serverless=_is_serverless_endpoint(self.config.platform.endpoint),
        )
        new_app_idl = app._to_union_idl(settings=settings)

        get_app_request = GetRequest(app_id=new_app_idl.metadata.id)
        get_response: GetResponse = self.sync_client.Get(get_app_request)

        old_app_idl = get_response.app

        initial_condition = None
        if old_app_idl.status.conditions:
            initial_condition = old_app_idl.status.conditions[-1]

        updated_app_idl = App._update_app_idl(old_app_idl, new_app_idl)

        def run_request():
            update_request = UpdateRequest(app=updated_app_idl)
            update_response: UpdateResponse = self.sync_client.Update(update_request)
            return update_response.app

        app_idl = self._watch_update(updated_app_idl, run_request, initial_condition)
        return app_idl

    def create(self, app: App, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        additional_distribution = self.upload_files(app)
        materialized_input_values = self.materialize_values(app, project=project, domain=domain)
        settings = AppSerializationSettings(
            org=self.org,
            project=project or self.default_project,
            domain=domain or self.default_domain,
            additional_distribution=additional_distribution,
            desired_state=Spec.DesiredState.DESIRED_STATE_STARTED,
            materialized_inputs=materialized_input_values,
            is_serverless=_is_serverless_endpoint(self.config.platform.endpoint),
        )
        app_idl = app._to_union_idl(settings=settings)

        def run_request():
            create_request = CreateRequest(app=app_idl)
            create_response: CreateResponse = self.sync_client.Create(create_request)
            return create_response.app

        self._watch_update(app_idl, run_request, None)
        return app_idl

    def _watch_update(
        self, app_idl: AppIDL, run_request: Callable[[], AppIDL], initial_condition: Optional[Condition] = None
    ) -> AppIDL:
        console = Console()

        # TODO: Revise when log streaming comes back
        # initial_datetime = None
        # if initial_condition:
        #     initial_datetime = initial_condition.last_transition_time.ToDatetime().replace(tzinfo=timezone.utc)

        name = app_idl.metadata.id.name
        _console_url = self.generate_console_url(app_idl)
        console_url = f"[link={_console_url}]{_console_url}[/link]"
        console.print(f"✨ Deploying Application: [link={_console_url}]{name}[/link]")
        console.print(f"🔎 Console URL: {console_url}")

        updated_idl = run_request()
        watch_request = WatchRequest(app_id=updated_idl.metadata.id)
        get_app_request = GetRequest(app_id=updated_idl.metadata.id)

        async def watch_for_status():
            response: WatchResponse
            async for response in self.async_client.Watch(watch_request):
                status = response.update_event.updated_app.status
                yield status

        async def poll_for_status():
            loop = asyncio.get_running_loop()
            while True:
                get_response: GetResponse = await loop.run_in_executor(None, self.sync_client.Get, get_app_request)
                yield get_response.app.status
                await asyncio.sleep(2)

        async def watch_status(console):
            latest_message = ""
            status: Status

            loop = asyncio.get_running_loop()
            async for status in merge(loop, watch_for_status(), poll_for_status()):
                if not status.conditions:
                    continue
                latest_condition = status.conditions[-1]
                if initial_condition == latest_condition:
                    # The status did not update from the initial condition
                    continue
                deployment_status = latest_condition.deployment_status
                status_str = Status.DeploymentStatus.Name(deployment_status).split("_")[-1].title()
                if latest_condition.message != latest_message:
                    console.print(f"[bold]\\[Status][/] [italic]{status_str}:[/] {latest_condition.message}")
                    latest_message = latest_condition.message

                if deployment_status in (
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_UNSPECIFIED,
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_STOPPED,
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_FAILED,
                ):
                    raise ClickException("Application failed to deploy")
                elif deployment_status == Status.DEPLOYMENT_STATUS_UNASSIGNED:
                    raise ClickException("Application was unassigned")
                elif deployment_status == Status.DeploymentStatus.DEPLOYMENT_STATUS_STARTED:
                    yield True

        async def watch(console):
            # TODO: Fix log streaming
            # status_iter = merge(watch_status(console), self._watch_logs(updated_idl, initial_datetime, console))
            status_iter = watch_status(console)
            try:
                async for to_stop in status_iter:
                    if to_stop:
                        return
            finally:
                await status_iter.aclose()

        run_sync(watch, console)

        _url = updated_idl.status.ingress.public_url
        url = f"[link={_url}]{_url}[/link]"

        console.print()
        console.print(f"🚀 Deployed Endpoint: {url}")
        return updated_idl

    @staticmethod
    def _format_line(line: str, initial_datetime: Optional[datetime]) -> str:
        whitespace_idx = line.find(" ")
        if whitespace_idx == -1:
            return line
        datetime_str = line[:whitespace_idx]

        try:
            dt = datetime.fromisoformat(datetime_str)
        except ValueError:
            return line

        if initial_datetime is not None and dt > initial_datetime:
            # Do not show logs that are older than the previous initial condition
            return ""

        formatted_dt = dt.strftime("%H:%M:%S")
        return f"{formatted_dt}{line[whitespace_idx:]}"

    async def _watch_logs(self, app_idl: AppIDL, initial_datetime: Optional[datetime], console: Console):
        async_log_client = AppLogsServiceStub(self.async_channel)
        tail_request = TailLogsRequest(app_id=app_idl.metadata.id)
        response: TailLogsResponse
        async for response in async_log_client.TailLogs(tail_request):
            for log in response.batches.logs:
                for line in log.lines:
                    formatted_line = self._format_line(line, initial_datetime).strip()
                    if not formatted_line:
                        continue
                    console.print(f"[bold]\\[App][/] {formatted_line}")
                    # TODO: Detect errors and stop

            yield False

    def stop(self, name: str, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        app_idl = self.get(name, project=project, domain=domain)
        app_idl.spec.desired_state = Spec.DesiredState.DESIRED_STATE_STOPPED

        console = Console()

        console_url = self.generate_console_url(app_idl)
        console.print(f"🛑 Stopping Application: [link={console_url}]{name}[/link]")

        update_request = UpdateRequest(app=app_idl)
        update_response: UpdateResponse = self.sync_client.Update(update_request)
        return update_response.app

    def upload_files(self, app: App) -> Optional[str]:
        """Upload files required by app."""
        if not app.include:
            return None

        with TemporaryDirectory() as temp_dir:
            tar_path = os.path.join(temp_dir, FILES_TAR_FILE_NAME)
            with tarfile.open(tar_path, "w:gz") as tar:
                for resolve_include in app.include_resolved:
                    tar.add(resolve_include.src, arcname=resolve_include.dest)

            _, upload_native_url = self.union_remote.upload_file(Path(tar_path))

            return upload_native_url

    def materialize_values(self, app: App, project: Optional[str] = None, domain: Optional[str] = None) -> dict:
        output = {}
        for user_input in app.inputs:
            if isinstance(user_input.value, ArtifactQuery):
                query = deepcopy(user_input.value)
                query.project = project or self.default_project
                query.domain = domain or self.default_domain

                result = self.union_remote.get_artifact(query=query.to_flyte_idl())
                scalar = result.literal.scalar
                input_type = None
                if scalar.blob is not None:
                    # Handle FlyteFiles and FlyteDirectories
                    if scalar.blob.metadata.type.dimensionality == BlobType.BlobDimensionality.SINGLE:
                        input_type = Input.Type.File
                    elif scalar.blob.metadata.type.dimensionality == BlobType.BlobDimensionality.MULTIPART:
                        input_type = Input.Type.Directory
                    value = result.literal.scalar.blob.uri
                elif scalar.primitive is not None and scalar.primitive.string_value is not None:
                    value = scalar.primitive.string_value
                    input_type = Input.Type.String
                else:
                    msg = (
                        f"Artifact '{result.name}' is not supported as an App input, Only strings, files, "
                        "and directory are supported."
                    )
                    raise ValueError(msg)

                output[user_input.name] = MaterializedInput(
                    value=value,
                    type=input_type,
                )

            elif isinstance(user_input.value, URLQuery):
                query = user_input.value
                # TODO: Assuming application has the same project and domain
                # TODO: Raise more informative error the assumption does not hold
                app_idl = self.get(name=query.name, project=project, domain=domain)
                if user_input.value.public:
                    output[user_input.name] = MaterializedInput(
                        value=app_idl.status.ingress.public_url,
                        type=Input.Type.String,
                    )
                else:
                    app_id = app_idl.metadata.id
                    output[user_input.name] = MaterializedInput(
                        value=app_id.name,
                        type=Input.Type._UrlQuery,
                    )

        return output

    def generate_console_url(self, app_idl: AppIDL) -> str:
        """Generate console url for app_idl."""
        http_domain = self.union_remote.generate_console_http_domain()
        org = _get_organization(self.config.platform, self.sync_channel)

        app_id = app_idl.metadata.id

        if org is None or org == "":
            console = "console"
        else:
            console = f"org/{org}"

        return f"{http_domain}/{console}/projects/{app_id.project}/domains/{app_id.domain}/apps/{app_id.name}"

    @property
    def sync_channel(self) -> grpc.Channel:
        try:
            return self._sync_channel
        except AttributeError:
            self._sync_channel, self._org = _get_channel_with_org(self.config.platform)
            return self._sync_channel

    @property
    def org(self) -> Optional[str]:
        try:
            return self._org
        except AttributeError:
            self._sync_channel, self._org = _get_channel_with_org(self.config.platform)
            if self._org is None or self._org == "":
                self._org = self.config.platform.endpoint.split(".")[0]
            return self._org

    @property
    def sync_client(self) -> AppServiceStub:
        try:
            return self._sync_client
        except AttributeError:
            self._sync_client = AppServiceStub(self.sync_channel)
            return self._sync_client

    @property
    def async_client(self) -> AppServiceStub:
        try:
            return self._async_client
        except AttributeError:
            self._async_client = AppServiceStub(self.async_channel)
            return self._async_client

    @property
    def async_channel(self) -> grpc.aio.Channel:
        from union.filesystems._endpoint import _create_secure_channel_from_config

        try:
            return self._async_channel
        except AttributeError:
            self._async_channel = _create_secure_channel_from_config(self.config.platform, self.sync_channel)
            return self._async_channel

    @staticmethod
    def deployment_status(app_idl: AppIDL) -> str:
        try:
            current_status = app_idl.status.conditions[-1].deployment_status
            return Status.DeploymentStatus.Name(current_status).split("_")[-1].title()
        except Exception:
            return "Unknown"

    @staticmethod
    def desired_state(app_idl: AppIDL) -> str:
        return Spec.DesiredState.Name(app_idl.spec.desired_state).split("_")[-1].title()

    @staticmethod
    def get_limits(app_idl: AppIDL) -> dict:
        output = {}
        for limit in app_idl.spec.container.resources.limits:
            if limit.name == Resources.ResourceName.CPU:
                output["cpu"] = limit.value
            if limit.name == Resources.ResourceName.MEMORY:
                output["memory"] = limit.value
        return output



import re
from datetime import timedelta


def _timedelta_format(td_object: timedelta) -> str:
    """Convert timedelta format to readable string"""
    seconds = int(td_object.total_seconds())
    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]

    strings = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            has_s = "s" if period_value > 1 else ""
            strings.append(f"{period_value} {period_name}{has_s}")

    return ", ".join(strings)


def _sanitize_label(label: str) -> str:
    """Sanitize name for Union labels."""
    return re.sub("[:/]", "_", label)



from union.internal.common import identity_pb2 as _identity_pb2
from flyteidl.core import identifier_pb2 as _identifier_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from union.internal.workspace import workspace_definition_pb2 as _workspace_definition_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class WorkspaceInstanceIdentifier(_message.Message):
    __slots__ = ["org", "project", "domain", "name", "host"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    name: str
    host: str
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ..., host: _Optional[str] = ...) -> None: ...

class WorkspaceInstanceStatus(_message.Message):
    __slots__ = ["phase"]
    class WorkspaceInstancePhase(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        WORKSPACE_INSTANCE_PHASE_UNSPECIFIED: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
        WORKSPACE_INSTANCE_PHASE_PENDING: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
        WORKSPACE_INSTANCE_PHASE_RUNNING: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
        WORKSPACE_INSTANCE_PHASE_TERMINATING: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
        WORKSPACE_INSTANCE_PHASE_TERMINATED: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
        WORKSPACE_INSTANCE_PHASE_FAILED: _ClassVar[WorkspaceInstanceStatus.WorkspaceInstancePhase]
    WORKSPACE_INSTANCE_PHASE_UNSPECIFIED: WorkspaceInstanceStatus.WorkspaceInstancePhase
    WORKSPACE_INSTANCE_PHASE_PENDING: WorkspaceInstanceStatus.WorkspaceInstancePhase
    WORKSPACE_INSTANCE_PHASE_RUNNING: WorkspaceInstanceStatus.WorkspaceInstancePhase
    WORKSPACE_INSTANCE_PHASE_TERMINATING: WorkspaceInstanceStatus.WorkspaceInstancePhase
    WORKSPACE_INSTANCE_PHASE_TERMINATED: WorkspaceInstanceStatus.WorkspaceInstancePhase
    WORKSPACE_INSTANCE_PHASE_FAILED: WorkspaceInstanceStatus.WorkspaceInstancePhase
    PHASE_FIELD_NUMBER: _ClassVar[int]
    phase: WorkspaceInstanceStatus.WorkspaceInstancePhase
    def __init__(self, phase: _Optional[_Union[WorkspaceInstanceStatus.WorkspaceInstancePhase, str]] = ...) -> None: ...

class WorkspaceInstanceSpec(_message.Message):
    __slots__ = ["status", "identity", "created_at", "updated_at", "workspace_definition_id", "workflow_execution_id", "uri"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    IDENTITY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_DEFINITION_ID_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    URI_FIELD_NUMBER: _ClassVar[int]
    status: WorkspaceInstanceStatus
    identity: _identity_pb2.Identity
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    workspace_definition_id: _workspace_definition_pb2.WorkspaceDefinitionIdentifier
    workflow_execution_id: _identifier_pb2.WorkflowExecutionIdentifier
    uri: str
    def __init__(self, status: _Optional[_Union[WorkspaceInstanceStatus, _Mapping]] = ..., identity: _Optional[_Union[_identity_pb2.Identity, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., workspace_definition_id: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionIdentifier, _Mapping]] = ..., workflow_execution_id: _Optional[_Union[_identifier_pb2.WorkflowExecutionIdentifier, _Mapping]] = ..., uri: _Optional[str] = ...) -> None: ...

class WorkspaceInstance(_message.Message):
    __slots__ = ["id", "spec"]
    ID_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    id: WorkspaceInstanceIdentifier
    spec: WorkspaceInstanceSpec
    def __init__(self, id: _Optional[_Union[WorkspaceInstanceIdentifier, _Mapping]] = ..., spec: _Optional[_Union[WorkspaceInstanceSpec, _Mapping]] = ...) -> None: ...

class WorkspaceInstances(_message.Message):
    __slots__ = ["instances"]
    INSTANCES_FIELD_NUMBER: _ClassVar[int]
    instances: _containers.RepeatedCompositeFieldContainer[WorkspaceInstance]
    def __init__(self, instances: _Optional[_Iterable[_Union[WorkspaceInstance, _Mapping]]] = ...) -> None: ...

class StartEvent(_message.Message):
    __slots__ = ["workspace"]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    workspace: WorkspaceInstance
    def __init__(self, workspace: _Optional[_Union[WorkspaceInstance, _Mapping]] = ...) -> None: ...

class UpdateEvent(_message.Message):
    __slots__ = ["updated_workspace", "old_workspace"]
    UPDATED_WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    OLD_WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    updated_workspace: WorkspaceInstance
    old_workspace: WorkspaceInstance
    def __init__(self, updated_workspace: _Optional[_Union[WorkspaceInstance, _Mapping]] = ..., old_workspace: _Optional[_Union[WorkspaceInstance, _Mapping]] = ...) -> None: ...

class StopEvent(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: WorkspaceInstanceIdentifier
    def __init__(self, id: _Optional[_Union[WorkspaceInstanceIdentifier, _Mapping]] = ...) -> None: ...

class StoppingEvent(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: WorkspaceInstanceIdentifier
    def __init__(self, id: _Optional[_Union[WorkspaceInstanceIdentifier, _Mapping]] = ...) -> None: ...



from contextlib import contextmanager

from flytekit.image_spec import default_builder

from union.ucimage import _image_builder


def _return_empty_str():
    return ""


@contextmanager
def patch_get_flytekit_for_pypi():
    """Mock out get_flytekit_for_pypi to return an empty string."""
    try:
        orig_get_flytekit_for_pypi = default_builder.get_flytekit_for_pypi
        default_builder.get_flytekit_for_pypi = _return_empty_str
        yield
        default_builder.get_flytekit_for_pypi = orig_get_flytekit_for_pypi
    except AttributeError:
        # Catch this error just in case `flytekit` changes the location of
        # get_flytekit_for_pypi
        yield


@contextmanager
def patch_get_unionai_for_pypi():
    """Mock out get_unionai_for_pypi to return an empty string."""
    orig_get_unionai_for_pypi = _image_builder.get_unionai_for_pypi
    _image_builder.get_unionai_for_pypi = _return_empty_str
    yield
    _image_builder.get_unionai_for_pypi = orig_get_unionai_for_pypi



import logging
import random
import string
from collections import namedtuple
from collections.abc import MutableMapping, Sequence
from contextlib import suppress
from typing import Any, Optional

import grpc
from flyteidl.service.admin_pb2_grpc import AdminServiceStub
from flyteidl.service.dataproxy_pb2_grpc import DataProxyServiceStub
from flyteidl.service.signal_pb2_grpc import SignalServiceStub
from flytekit.clients.friendly import SynchronousFlyteClient

logger = logging.getLogger(__name__)


class ConcreteClientCallDetails(
    namedtuple(
        "ConcreteClientCallDetails",
        (
            "method",
            "timeout",
            "metadata",
            "credentials",
            "wait_for_ready",
            "compression",
        ),
    ),
    grpc.ClientCallDetails,
):
    """Concrete implementation for ClientCallDetails."""


def set_org(obj: Any, org: str):
    """Recursively set the org in obj."""
    with suppress(AttributeError):
        obj.org = org

    with suppress(AttributeError):
        for _, value in obj.ListFields():
            set_org(value, org=org)

    if isinstance(obj, Sequence) and not isinstance(obj, str):
        for value in obj:
            set_org(value, org=org)

    if isinstance(obj, MutableMapping):
        for value in obj.values():
            set_org(value, org=org)


def generate_random_str(N):
    return "".join(random.choice(string.ascii_letters) for _ in range(N))


class OrgInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor that adds org metadata into the grpc call."""

    def __init__(self, org: str):
        self.org = org

    def intercept_unary_unary(self, continuation, client_call_details, request):
        request_id = f"u-{generate_random_str(20)}"
        old_metadata = client_call_details.metadata or []
        new_metadata = old_metadata + [("x-user-active-org", self.org), ("x-request-id", request_id)]
        new_details = ConcreteClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=new_metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
            compression=client_call_details.compression,
        )

        # Inject org into every IDL requests
        set_org(request, org=self.org)
        logger.debug(f"request: {type(request)}, x-request-id: {request_id}, contents:\n{request}")
        return continuation(new_details, request)


def update_client_with_interceptor(client: SynchronousFlyteClient, org: Optional[str]) -> SynchronousFlyteClient:
    """Updates SynchronousFlyteClient inplace with an interceptor with org."""
    if org is None:
        # There is no org, so no need to intercept
        return client

    # Intercept calls to FlyteRemote and add org to grpc metadata
    org_interceptor = OrgInterceptor(org=org)
    previous_channel = client._channel
    new_channel = grpc.intercept_channel(previous_channel, org_interceptor)

    client._channel = new_channel
    client._stub = AdminServiceStub(new_channel)
    client._signal = SignalServiceStub(new_channel)
    client._dataproxy_stub = DataProxyServiceStub(new_channel)

    return client


def intercept_channel_with_org(org: Optional[str], channel: grpc.Channel):
    if org is None:
        return channel

    org_interceptor = OrgInterceptor(org=org)
    return grpc.intercept_channel(channel, org_interceptor)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common/identifier.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x63ommon/identifier.proto\x12\x0f\x63loudidl.common\x1a\x17validate/validate.proto\"~\n\x11ProjectIdentifier\x12+\n\x0corganization\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x0corganization\x12\x1f\n\x06\x64omain\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x06\x64omain\x12\x1b\n\x04name\x18\x03 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04name\"T\n\x11\x43lusterIdentifier\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x1b\n\x04name\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04name\"O\n\x15\x43lusterPoolIdentifier\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\"_\n\x17\x43lusterConfigIdentifier\x12+\n\x0corganization\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x0corganization\x12\x17\n\x02id\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x02id\"3\n\x0eUserIdentifier\x12!\n\x07subject\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x07subject\":\n\x15\x41pplicationIdentifier\x12!\n\x07subject\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x07subject\"Q\n\x0eRoleIdentifier\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x1b\n\x04name\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04name\"O\n\rOrgIdentifier\x12>\n\x04name\x18\x01 \x01(\tB*\xfa\x42\'r%\x10\x01\x18?2\x1f^[a-z0-9]([-a-z0-9]*[a-z0-9])?$R\x04name\"y\n\x18ManagedClusterIdentifier\x12\x1b\n\x04name\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04name\x12:\n\x03org\x18\x03 \x01(\x0b\x32\x1e.cloudidl.common.OrgIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x03orgJ\x04\x08\x01\x10\x02\"S\n\x10PolicyIdentifier\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x1b\n\x04name\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04nameB\xb0\x01\n\x13\x63om.cloudidl.commonB\x0fIdentifierProtoH\x02P\x01Z)github.com/unionai/cloud/gen/pb-go/common\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43loudidl.Common\xca\x02\x0f\x43loudidl\\Common\xe2\x02\x1b\x43loudidl\\Common\\GPBMetadata\xea\x02\x10\x43loudidl::Commonb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common.identifier_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.cloudidl.commonB\017IdentifierProtoH\002P\001Z)github.com/unionai/cloud/gen/pb-go/common\242\002\003CCX\252\002\017Cloudidl.Common\312\002\017Cloudidl\\Common\342\002\033Cloudidl\\Common\\GPBMetadata\352\002\020Cloudidl::Common'
  _PROJECTIDENTIFIER.fields_by_name['organization']._options = None
  _PROJECTIDENTIFIER.fields_by_name['organization']._serialized_options = b'\372B\004r\002\020\001'
  _PROJECTIDENTIFIER.fields_by_name['domain']._options = None
  _PROJECTIDENTIFIER.fields_by_name['domain']._serialized_options = b'\372B\004r\002\020\001'
  _PROJECTIDENTIFIER.fields_by_name['name']._options = None
  _PROJECTIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _CLUSTERIDENTIFIER.fields_by_name['name']._options = None
  _CLUSTERIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _CLUSTERCONFIGIDENTIFIER.fields_by_name['organization']._options = None
  _CLUSTERCONFIGIDENTIFIER.fields_by_name['organization']._serialized_options = b'\372B\004r\002\020\001'
  _CLUSTERCONFIGIDENTIFIER.fields_by_name['id']._options = None
  _CLUSTERCONFIGIDENTIFIER.fields_by_name['id']._serialized_options = b'\372B\004r\002\020\001'
  _USERIDENTIFIER.fields_by_name['subject']._options = None
  _USERIDENTIFIER.fields_by_name['subject']._serialized_options = b'\372B\004r\002\020\001'
  _APPLICATIONIDENTIFIER.fields_by_name['subject']._options = None
  _APPLICATIONIDENTIFIER.fields_by_name['subject']._serialized_options = b'\372B\004r\002\020\001'
  _ROLEIDENTIFIER.fields_by_name['name']._options = None
  _ROLEIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _ORGIDENTIFIER.fields_by_name['name']._options = None
  _ORGIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\'r%\020\001\030?2\037^[a-z0-9]([-a-z0-9]*[a-z0-9])?$'
  _MANAGEDCLUSTERIDENTIFIER.fields_by_name['name']._options = None
  _MANAGEDCLUSTERIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _MANAGEDCLUSTERIDENTIFIER.fields_by_name['org']._options = None
  _MANAGEDCLUSTERIDENTIFIER.fields_by_name['org']._serialized_options = b'\372B\005\212\001\002\020\001'
  _POLICYIDENTIFIER.fields_by_name['name']._options = None
  _POLICYIDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _globals['_PROJECTIDENTIFIER']._serialized_start=69
  _globals['_PROJECTIDENTIFIER']._serialized_end=195
  _globals['_CLUSTERIDENTIFIER']._serialized_start=197
  _globals['_CLUSTERIDENTIFIER']._serialized_end=281
  _globals['_CLUSTERPOOLIDENTIFIER']._serialized_start=283
  _globals['_CLUSTERPOOLIDENTIFIER']._serialized_end=362
  _globals['_CLUSTERCONFIGIDENTIFIER']._serialized_start=364
  _globals['_CLUSTERCONFIGIDENTIFIER']._serialized_end=459
  _globals['_USERIDENTIFIER']._serialized_start=461
  _globals['_USERIDENTIFIER']._serialized_end=512
  _globals['_APPLICATIONIDENTIFIER']._serialized_start=514
  _globals['_APPLICATIONIDENTIFIER']._serialized_end=572
  _globals['_ROLEIDENTIFIER']._serialized_start=574
  _globals['_ROLEIDENTIFIER']._serialized_end=655
  _globals['_ORGIDENTIFIER']._serialized_start=657
  _globals['_ORGIDENTIFIER']._serialized_end=736
  _globals['_MANAGEDCLUSTERIDENTIFIER']._serialized_start=738
  _globals['_MANAGEDCLUSTERIDENTIFIER']._serialized_end=859
  _globals['_POLICYIDENTIFIER']._serialized_start=861
  _globals['_POLICYIDENTIFIER']._serialized_end=944
# @@protoc_insertion_point(module_scope)



from typing import Optional

import rich_click as click
from click import secho, style
from flytekit.clis.sdk_in_container.constants import CTX_VERBOSE
from flytekit.models.task import Resources
from grpc import RpcError
from rich.console import Console
from rich.table import Table

from union._config import _DEFAULT_DOMAIN, _DEFAULT_PROJECT_BYOC, _get_config_obj
from union.cli._common import _get_channel_with_org
from union.internal.common.list_pb2 import ListRequest
from union.internal.identity.app_payload_pb2 import ListAppsRequest
from union.internal.identity.app_service_pb2_grpc import AppsServiceStub
from union.internal.secret.payload_pb2 import ListSecretsRequest
from union.internal.secret.secret_pb2_grpc import SecretServiceStub
from union.remote import UnionRemote


@click.group()
def get():
    """Get a resource."""


@get.command()
@click.pass_context
@click.option("--project", help="Project name")
@click.option("--domain", help="Domain name")
def secret(
    ctx: click.Context,
    project: Optional[str],
    domain: Optional[str],
):
    """Get secrets."""
    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    stub = SecretServiceStub(channel)
    secrets = []

    next_token, has_next = "", True

    try:
        while has_next:
            request = ListSecretsRequest(domain=domain, project=project, limit=20, token=next_token, organization=org)
            response = stub.ListSecrets(request)
            next_token = response.token
            has_next = next_token != ""

            secrets.extend(response.secrets)
    except RpcError as e:
        raise click.ClickException(f"Unable to get secrets.\n{e}") from e

    if secrets:
        table = Table()
        for name in ["name", "project", "domain"]:
            table.add_column(name, justify="right")

        for secret in secrets:
            project = secret.id.project or "-"
            domain = secret.id.domain or "-"
            table.add_row(secret.id.name, project, domain)

        console = Console()
        console.print(table)
    else:
        click.echo("No secrets found")


@get.group("api-key")
def api_key():
    """Manage API keys."""


@api_key.command("admin")
def admin():
    """Show existing API keys for admin."""
    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    stub = AppsServiceStub(channel)

    apps = []
    next_token, has_next = "", True
    try:
        while has_next:
            request = ListAppsRequest(organization=org, request=ListRequest(limit=20, token=next_token))
            response = stub.List(request)
            next_token = response.token
            has_next = next_token != ""

            apps.extend(response.apps)
    except RpcError as e:
        raise click.ClickException(f"Unable to get apps.\n{e}") from e

    if apps:
        table = Table()
        table.add_column("name", overflow="fold")

        for app in apps:
            table.add_row(app.client_id)

        console = Console()
        console.print(table)
    else:
        click.echo("No apps found.")


@get.command()
@click.pass_context
@click.option("--name", default=None)
@click.option("--project", default=_DEFAULT_PROJECT_BYOC, help="Project name")
@click.option("--domain", default=_DEFAULT_DOMAIN, help="Domain name")
def apps(ctx: click.Context, name: Optional[str], project: str, domain: str):
    """Get apps."""
    verbose_on = ctx.obj[CTX_VERBOSE] > 0
    from union.configuration import UnionAIPlugin

    remote = UnionAIPlugin.get_remote(config=None, project=project, domain=domain)
    app_remote = remote._app_remote

    if name is None:
        # TODO: Add filter
        app_list = app_remote.list()
        if verbose_on:
            secho(app_list)
            return

        if not app_list:
            secho("No applications running")
            return

        table = Table()
        table.add_column("Name", overflow="fold")
        table.add_column("Link", overflow="fold")
        table.add_column("Status", overflow="fold")
        table.add_column("Desired State", overflow="fold")
        table.add_column("CPU", overflow="fold")
        table.add_column("Memory", overflow="fold")

        table_items = []

        for app in app_list:
            url = f"[link={app.status.ingress.public_url}]Click Here[/link]"
            limits = app_remote.get_limits(app)

            status = app_remote.deployment_status(app)
            desired_state = app_remote.desired_state(app)

            table_items.append(
                {
                    "name": app.metadata.id.name,
                    "url": url,
                    "status": status,
                    "desired_state": desired_state,
                    "cpu": limits.get("cpu", "-"),
                    "memory": limits.get("memory", "-"),
                }
            )

        sorted_table_items = sorted(table_items, key=lambda item: item["status"])

        for item in sorted_table_items:
            table.add_row(
                item["name"],
                item["url"],
                item["status"],
                item["desired_state"],
                item["cpu"],
                item["memory"],
            )

        console = Console()
        console.print(table)
    else:
        app_idl = app_remote.get(name=name)
        if verbose_on:
            secho(app_idl)
        else:
            limits = app_remote.get_limits(app_idl)
            details = {
                "name": app_idl.metadata.id.name,
                "status": app_remote.deployment_status(app_idl),
                "link": app_idl.status.ingress.public_url,
                "desired_state": app_remote.desired_state(app_idl),
                "cpu": limits.get("cpu", "-"),
                "memory": limits.get("memory", "-"),
            }
            items = ["name", "link", "status", "cpu", "memory", "desired_state"]
            max_item_len = max(len(i) for i in items)

            for item in items:
                styled_name = style(f"{item:<{max_item_len}}", bold=True)
                value = details[item]
                secho(f"{styled_name} : {value}")


@get.command()
@click.pass_context
@click.option("--name", default=None)
@click.option("--project", default=_DEFAULT_PROJECT_BYOC, help="Project name")
@click.option("--domain", default=_DEFAULT_DOMAIN, help="Domain name")
@click.option("--show-details", default=False, is_flag=True, help="Show additional details")
def workspace(ctx: click.Context, name: Optional[str], project: str, domain: str, show_details: bool):
    """Get workspaces."""

    from union._async import run_sync
    from union.internal.workspace.workspace_definition_payload_pb2 import ListWorkspaceDefinitionsResponse
    from union.workspace._vscode_remote import WorkspaceRemote

    remote = UnionRemote(default_domain=domain, default_project=project)
    ws_remote = WorkspaceRemote(default_domain=domain, default_project=project, union_remote=remote)

    ws_defs: ListWorkspaceDefinitionsResponse = run_sync(
        ws_remote.list_workspace_definitions,
        name,
        sort_by="created_at",
        direction="desc",
    )
    if len(ws_defs.workspace_definitions) == 0:
        click.echo("No workspaces found.")
        return

    table = Table(header_style="yellow", border_style="yellow")
    table.add_column("Workspace name", overflow="fold")
    table.add_column("CPU", overflow="fold")
    table.add_column("Memory", overflow="fold")
    table.add_column("GPU", overflow="fold")
    table.add_column("Accelerator", overflow="fold")
    table.add_column("TTL Seconds", overflow="fold")
    table.add_column("Active URL", overflow="fold", style="green")

    workspace_names = set()
    for ws_def in ws_defs.workspace_definitions:
        if ws_def.id.name in workspace_names:
            continue

        workspace_names.add(ws_def.id.name)
        try:
            uri = run_sync(ws_remote.get_latest_workspace_instance_uri, ws_def.id.name, ws_def.id.version)
            if uri is None:
                loading_url = None
            else:
                url = f"https://{remote.config.platform.endpoint}/{uri}"
                loading_url = f"https://{remote.config.platform.endpoint}/loading?type=workspace&url={url}"
        except RuntimeError:
            loading_url = None

        cpu, memory, gpu = None, None, None
        resources = Resources.from_flyte_idl(ws_def.spec.resources)
        for i in range(len(resources.requests)):
            name = resources.requests[i].name
            value = resources.requests[i].value

            if name == Resources.ResourceName.CPU:
                cpu = value
            elif name == Resources.ResourceName.MEMORY:
                memory = value
            elif name == Resources.ResourceName.GPU:
                gpu = value

        table.add_row(
            ws_def.id.name,
            cpu or "-",
            memory or "-",
            gpu or "-",
            str(ws_def.spec.extended_resources.gpu_accelerator) or "-",
            str(ws_def.spec.ttl_seconds),
            f"[link={loading_url}]Open in Browser[/link]" if loading_url else "-",
        )

    console = Console()
    console.print(table)


@get.command()
@click.argument("uri", metavar="URI of the type flyte://av0/...")
@click.pass_context
def artifact(ctx: click.Context, uri: str):
    """Get artifacts."""
    remote = UnionRemote()
    art = remote.get_artifact(uri=uri)
    url = remote.generate_console_url(art)
    Console().print(
        f"[green]Artifact: [bold][link={url}]{art.name}[/link][/bold]"
        f" and Artifact ID: [bold]{art.metadata().uri}[/bold][/green]"
    )  # noqa



from __future__ import annotations

import datetime
import logging
import os
import typing
from contextlib import contextmanager, suppress
from dataclasses import dataclass
from typing import List, Optional, Union

from flyteidl.core import artifact_id_pb2 as art_id
from flyteidl.service.identity_pb2 import UserInfoRequest, UserInfoResponse
from flyteidl.service.identity_pb2_grpc import IdentityServiceStub
from flytekit import BlobType, ImageSpec, Resources
from flytekit.clients.friendly import SynchronousFlyteClient
from flytekit.configuration import Config
from flytekit.core.artifact import ArtifactQuery, Partitions, TimePartition
from flytekit.core.type_engine import LiteralsResolver, TypeEngine
from flytekit.exceptions import user as user_exceptions
from flytekit.exceptions.user import FlyteEntityNotExistException, FlyteInvalidInputException
from flytekit.models.literals import Blob, BlobMetadata, Literal, Scalar
from flytekit.remote import FlyteRemote
from flytekit.remote.entities import FlyteLaunchPlan, FlyteTask, FlyteWorkflow
from flytekit.remote.executions import FlyteNodeExecution, FlyteTaskExecution, FlyteWorkflowExecution
from flytekit.tools.interactive import ipython_check
from flytekit.tools.translator import Options

from union._config import (
    _DEFAULT_DOMAIN,
    _DEFAULT_PROJECT_BYOC,
    ConfigSource,
    ConfigWithSource,
    _config_from_api_key,
    _get_config_obj,
    _get_default_project,
    _get_organization,
)
from union._interceptor import update_client_with_interceptor
from union.app._models import App
from union.artifacts import Artifact
from union.artifacts._utils import construct_search_artifact_request
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.artifacts import artifacts_pb2, artifacts_pb2_grpc
from union.internal.identity import user_service_pb2_grpc
from union.internal.imagebuilder import definition_pb2 as image_definition__pb2
from union.internal.imagebuilder import payload_pb2 as image_payload__pb2
from union.internal.imagebuilder import service_pb2_grpc as image_service_pb2_grpc
from union.internal.secret.definition_pb2 import Secret
from union.internal.secret.payload_pb2 import ListSecretsRequest, ListSecretsResponse
from union.internal.secret.secret_pb2_grpc import SecretServiceStub
from union.remote._app_template_factory import AppTemplate, HuggingFaceModelInfo, get_app_templates_for_model

SERVERLESS_VANITY_URLS = {
    "https://serverless-1.us-east-2.s.union.ai": "https://serverless.union.ai",
    "https://serverless-preview.canary.unionai.cloud": "https://serverless.canary.union.ai",
    "https://serverless-gcp.cloud-staging.union.ai": "https://serverless.staging.union.ai",
}


@dataclass
class DeployableModel:
    artifact: Artifact
    app_templates: List[AppTemplate]


_hf_download_image = ImageSpec(
    name="hfhub-cache",
    registry="ghcr.io/unionai-oss",
    packages=["huggingface_hub[hf_transfer]==0.26.3", "union==0.1.150"],
)


class UnionRemote(FlyteRemote):
    def __init__(
        self,
        config: typing.Optional[Union[Config, str]] = None,
        default_project: typing.Optional[str] = None,
        default_domain: typing.Optional[str] = None,
        data_upload_location: str = "flyte://my-s3-bucket/",
        interactive_mode_enabled: typing.Optional[bool] = None,
        **kwargs,
    ):
        from union._config import _get_image_builder_priority
        from union.ucimage._image_builder import _register_union_image_builder

        if config is None:
            config = _get_config_obj(config, default_to_union_semantics=True)
        else:
            config_with_source = ConfigWithSource(config=config, source=ConfigSource.REMOTE)
            config = _get_config_obj(config_with_source, default_to_union_semantics=True)

        # Creating a UnionRemote within a task uses the execution's project and domain
        # We use the internal environment variables here, so that it works without a flytekit entrypoint. For example,
        # serving does not use a flyte entrypoint.
        if default_project is None:
            default_project = os.getenv("FLYTE_INTERNAL_EXECUTION_PROJECT", None)
            if default_project is None:
                default_project = _get_default_project(_DEFAULT_PROJECT_BYOC, cfg_obj=config)

        if default_domain is None:
            default_domain = os.getenv("FLYTE_INTERNAL_EXECUTION_DOMAIN", _DEFAULT_DOMAIN)

        if default_domain is None:
            default_domain = _DEFAULT_DOMAIN

        # register Union image builder when getting remote so it's available for
        # jupyter notebook task and workflow execution.
        _register_union_image_builder(_get_image_builder_priority(config.platform.endpoint))

        # We set `interactive_mode_enabled=False` here and then adjust interactive_mode_enabled
        # mode afterwards to avoid the "Jupyter notebook and interactive task support is still alpha."
        # warning
        super().__init__(
            config=config,
            default_project=default_project,
            default_domain=default_domain,
            data_upload_location=data_upload_location,
            interactive_mode_enabled=False,
            **kwargs,
        )
        if interactive_mode_enabled is None:
            interactive_mode_enabled = ipython_check()
        self._interactive_mode_enabled = interactive_mode_enabled

        self._artifacts_client = None
        self._images_client = None
        self._secret_client = None
        self._users_client = None

        from union.remote._app_remote import AppRemote

        self._app_remote = AppRemote(union_remote=self, default_project=default_project, default_domain=default_domain)

    @classmethod
    def from_api_key(
        cls,
        api_key: str,
        default_project: typing.Optional[str] = None,
        default_domain: typing.Optional[str] = _DEFAULT_DOMAIN,
        data_upload_location: str = "flyte://my-s3-bucket/",
        **kwargs,
    ) -> "UnionRemote":
        """Call this if you want to directly instantiate a UnionRemote from an API key"""
        return cls(
            config=_config_from_api_key(api_key),
            default_project=default_project,
            default_domain=default_domain,
            data_upload_location=data_upload_location,
            **kwargs,
        )

    @property
    def client(self) -> SynchronousFlyteClient:
        """Return a SynchronousFlyteClient for additional operations."""
        if not self._client_initialized:
            client = SynchronousFlyteClient(self.config.platform, **self._kwargs)
            org = _get_organization(self.config.platform, channel=client._channel)
            self._client = update_client_with_interceptor(client, org)
            self._client_initialized = True

        return self._client

    def generate_console_http_domain(self) -> str:
        default_console_http_domain = super().generate_console_http_domain()
        return SERVERLESS_VANITY_URLS.get(default_console_http_domain, default_console_http_domain)

    def generate_console_url(
        self,
        entity: typing.Union[
            FlyteWorkflowExecution,
            FlyteNodeExecution,
            FlyteTaskExecution,
            FlyteWorkflow,
            FlyteTask,
            FlyteLaunchPlan,
            Artifact,
        ],
    ):
        """
        Generate a UnionAI console URL for the given Flyte remote endpoint.
        It will also handle Union AI specific entities like Artifacts.

        This will automatically determine if this is an execution or an entity
        and change the type automatically.
        """
        org = _get_organization(self.config.platform)
        if isinstance(entity, Artifact):
            url = f"{self.generate_console_http_domain()}/console/projects/{entity.project}/domains/{entity.domain}/artifacts/{entity.name}/versions/{entity.version}"  # noqa: E501
        else:
            url = super().generate_console_url(entity)
        if org is None:
            return url

        console_http_domain = self.generate_console_http_domain()
        old_prefix = f"{console_http_domain}/console/"
        new_prefix = f"{console_http_domain}/org/{org}/"

        return url.replace(old_prefix, new_prefix)

    @property
    def artifacts_client(self) -> artifacts_pb2_grpc.ArtifactRegistryStub:
        if self._artifacts_client:
            return self._artifacts_client

        self._artifacts_client = artifacts_pb2_grpc.ArtifactRegistryStub(self.client._channel)
        return self._artifacts_client

    @property
    def users_client(self) -> user_service_pb2_grpc.UserServiceStub:
        if self._users_client:
            return self._users_client

        self._users_client = user_service_pb2_grpc.UserServiceStub(self.client._channel)
        return self._users_client

    def search_artifacts(
        self,
        project: typing.Optional[str] = None,
        domain: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        artifact_key: typing.Optional[art_id.ArtifactKey] = None,
        query: typing.Optional[ArtifactQuery] = None,
        partitions: typing.Optional[Union[Partitions, typing.Dict[str, str]]] = None,
        time_partition: typing.Optional[Union[datetime.datetime, TimePartition]] = None,
        group_by_key: bool = False,
        limit: int = 100,
    ) -> typing.List[Artifact]:
        if limit > 500:
            raise ValueError("Limit cannot exceed 500")

        union_artifacts = []
        next_token = None
        while len(union_artifacts) < limit:
            search_request = construct_search_artifact_request(
                project=project or self.default_project,
                domain=domain or self.default_domain,
                name=name,
                artifact_key=artifact_key,
                query=query,
                partitions=partitions,
                time_partition=time_partition,
                group_by_key=group_by_key,
                limit=100,
                token=next_token,
            )
            search_response = self.artifacts_client.SearchArtifacts(search_request)
            artifact_list = search_response.artifacts
            next_token = search_response.token

            if len(artifact_list) == 0:
                break

            for al in artifact_list:
                ua = Artifact.from_flyte_idl(al)
                # assigned here because the resolver has an implicit dependency on the remote's context
                # can move into artifact if we are okay just using the current context.
                ua.resolver = LiteralsResolver(literals={ua.name: ua.literal}, variable_map=None, ctx=self.context)
                union_artifacts.append(ua)

        return union_artifacts

    def get_artifact(
        self,
        uri: typing.Optional[str] = None,
        artifact_key: typing.Optional[art_id.ArtifactKey] = None,
        artifact_id: typing.Optional[art_id.ArtifactID] = None,
        query: typing.Optional[typing.Union[art_id.ArtifactQuery, ArtifactQuery]] = None,
        get_details: bool = False,
    ) -> typing.Optional[Artifact]:
        """
        Get the specified artifact.

        :param uri: An artifact URI.
        :param artifact_key: An artifact key.
        :param artifact_id: The artifact ID.
        :param query: An artifact query.
        :param get_details: A bool to indicate whether or not to return artifact details.
        :return: The artifact as persisted in the service.
        """
        if query:
            if isinstance(query, ArtifactQuery):
                q = query.to_flyte_idl()
            else:
                q = query
        elif uri:
            q = art_id.ArtifactQuery(uri=uri)
        elif artifact_key:
            q = art_id.ArtifactQuery(artifact_id=art_id.ArtifactID(artifact_key=artifact_key))
        elif artifact_id:
            q = art_id.ArtifactQuery(artifact_id=artifact_id)
        else:
            raise ValueError("One of uri, key, id")
        req = artifacts_pb2.GetArtifactRequest(query=q, details=get_details)
        resp = self.artifacts_client.GetArtifact(req)
        a = Artifact.from_flyte_idl(resp.artifact)
        if a.literal and a.name:
            # assigned here because the resolver has an implicit dependency on the remote's context
            # can move into artifact if we are okay just using the current context.
            a.set_resolver(LiteralsResolver(literals={a.name: a.literal}, variable_map=None, ctx=self.context))
        return a

    def _execute(
        self,
        entity: typing.Union[FlyteTask, FlyteWorkflow, FlyteLaunchPlan],
        inputs: typing.Dict[str, typing.Any],
        project: str = None,
        domain: str = None,
        execution_name: typing.Optional[str] = None,
        execution_name_prefix: typing.Optional[str] = None,
        options: typing.Optional[Options] = None,
        wait: bool = False,
        type_hints: typing.Optional[typing.Dict[str, typing.Type]] = None,
        overwrite_cache: typing.Optional[bool] = None,
        envs: typing.Optional[typing.Dict[str, str]] = None,
        tags: typing.Optional[typing.List[str]] = None,
        cluster_pool: typing.Optional[str] = None,
        **kwargs,
    ) -> FlyteWorkflowExecution:
        resolved_inputs = {}
        for k, v in inputs.items():
            if isinstance(v, artifacts_pb2.Artifact):
                lit = v.spec.value
                resolved_inputs[k] = lit
            elif isinstance(v, Artifact):
                if v.literal is not None:
                    lit = v.literal
                elif v.to_id_idl() is not None:
                    fetched_artifact = self.get_artifact(artifact_id=v.concrete_artifact_id)
                    if not fetched_artifact:
                        raise user_exceptions.FlyteValueException(
                            v.concrete_artifact_id, "Could not find artifact with ID given"
                        )
                    lit = fetched_artifact.literal
                else:
                    raise user_exceptions.FlyteValueException(
                        v, "When binding input to Artifact, either the Literal or the ID must be set"
                    )
                resolved_inputs[k] = lit
            else:
                resolved_inputs[k] = v

        return super()._execute(
            entity,
            resolved_inputs,
            project=project,
            domain=domain,
            execution_name=execution_name,
            execution_name_prefix=execution_name_prefix,
            options=options,
            wait=wait,
            type_hints=type_hints,
            overwrite_cache=overwrite_cache,
            envs=envs,
            tags=tags,
            cluster_pool=cluster_pool,
            **kwargs,
        )

    def create_artifact(self, artifact: Artifact) -> Artifact:
        """
        Create an artifact in FlyteAdmin.

        :param artifact: The artifact to create.
        :return: The artifact as persisted in the service.
        """
        # Two things can happen here -
        #  - the call to to_literal may upload something, in the case of an offloaded data type.
        #    - if this happens, the upload request should already return the created Artifact object.
        if artifact.literal is None:
            with self.remote_context() as ctx:
                lt = artifact.literal_type or TypeEngine.to_literal_type(artifact.python_type)
                lit = TypeEngine.to_literal(ctx, artifact.python_val, artifact.python_type, lt)
                artifact.literal_type = lt.to_flyte_idl()
                if artifact.card:
                    card_key, card_val = artifact.card.serialize_to_string(ctx, artifact.name)
                    lit.set_metadata({card_key: card_val})
                artifact.literal = lit.to_flyte_idl()
        else:
            if artifact.literal_type is None:
                raise ValueError("Cannot create an artifact without a literal type set and a literal value present.")

        if artifact.project is None:
            artifact.project = self.default_project
        if artifact.domain is None:
            artifact.domain = self.default_domain
        create_request = Artifact.to_create_request(artifact)
        logging.debug(f"CreateArtifact request {create_request}")
        resp = self.artifacts_client.CreateArtifact(create_request)
        logging.debug(f"CreateArtifact response {resp}")
        # TODO artifact does not de-dupe on version, simply creates a new one, even for exact same partitions and data
        return Artifact.from_flyte_idl(resp.artifact)

    def deploy_app(self, app: App, project: Optional[str] = None, domain: Optional[str] = None) -> AppIDL:
        """
        Deploy an application.

        :param app: Application to deploy.
        :param project: Project name. If None, uses default_project.
        :param project: Domain name. If None, uses default_domain.
        :return: The App IDL for the deployed application.
        """
        return self._app_remote.deploy(app, project=project, domain=domain)

    def stop_app(self, name: str, project: Optional[str] = None, domain: Optional[str] = None):
        """
        Stop an application.

        :param name: Name of application to stop.
        :param project: Project name. If None, uses default_project.
        :param project: Domain name. If None, uses default_domain.
        :return: The App IDL for the stopped application.
        """
        return self._app_remote.stop(name=name, project=project, domain=domain)

    def _has_file_extension(self, file_path) -> bool:
        _, ext = os.path.splitext(file_path)
        return bool(ext)

    def get(self, uri: typing.Optional[str] = None) -> typing.Optional[typing.Union[LiteralsResolver, Literal, bytes]]:
        if not uri.startswith("union://"):
            # Assume this is the default behavior and this function is being called with a flyte uri
            return super().get(uri)
        if self._has_file_extension(uri):
            blob_dimensionality = BlobType.BlobDimensionality.SINGLE
        else:
            blob_dimensionality = BlobType.BlobDimensionality.MULTIPART
        return Literal(
            scalar=Scalar(
                blob=Blob(
                    metadata=BlobMetadata(
                        type=BlobType(
                            format="",
                            dimensionality=blob_dimensionality,
                        ),
                    ),
                    uri=uri,
                )
            )
        )

    def _create_model_from_hf(
        self,
        *,
        info: HuggingFaceModelInfo,
        hf_token_key: str,
        union_api_key: str,
        project: typing.Optional[str] = None,
        domain: typing.Optional[str] = None,
        chunk_size: int | None = None,
        retry: int = 0,
        resources: Optional[Resources] = None,
    ) -> FlyteWorkflowExecution:
        """
        This executes a background task to cache a model from Hugging Face and returns the handle to the execution.
        TODO maybe we should verify all requirements are met - 1. validate model 2. validate token exists
        and launch the execution with name=commit
        :param info: The model info.
        :param project: The project where the model will be stored.
        :param domain: The domain where the model will be stored.
        :param chunk_size: The chunk size to use when streaming the model files.
        :param retry: This can be used to force a new artifact to be created with the same name and an incremented
            version,
        :param resources: Specify compute resource requests for your task.
        :return: The model artifact.
        """
        if project is None:
            project = self.default_project
        if domain is None:
            domain = self.default_domain

        # Go through all the scopes to get secrets. There are tenants that do not org scoped secrets, so we have
        # iterate through the different scopes.
        secrets = (
            self._get_secrets() + self._get_secrets(project=project) + self._get_secrets(project=project, domain=domain)
        )

        if not self._secret_exists(secrets, hf_token_key, project, domain):
            msg = f"Secret named: {hf_token_key} does not exist. Please create one with `union create secret`"
            raise ValueError(msg)

        if not self._secret_exists(secrets, union_api_key, project, domain):
            msg = f"Secret named: {union_api_key} does not exist. Please create one with `union create secret`"
            raise ValueError(msg)

        from union.remote._cache_model import create_hf_model_cache_workflow

        image = self._get_hf_hub_download_image()
        model_cache_wf, source_path, additional_context = create_hf_model_cache_workflow(
            image, hf_token_key, union_api_key, resources=resources
        )

        mod_name = ".".join(model_cache_wf.name.split(".")[:-1])

        # TODO: Once with_overrides supports `secret_requests` in the model_cache_wf defined in
        # create_hf_model_cache_workflow, with a workflow using `with_overrides` on
        # requests and secret_requests. and remove this dynamic patch.
        with self._patch_version_from_hash(additional_context):
            wf = self.register_script(model_cache_wf, module_name=mod_name, source_path=source_path)

        inputs = {"info": info}
        inputs["chunk_size"] = chunk_size or 8 * 1024 * 1024
        inputs["retry"] = retry or 0
        inputs["hf_token_key"] = hf_token_key

        return self.execute(wf, inputs=inputs, project=project, domain=domain)

    @contextmanager
    def _patch_version_from_hash(self, additional_context: list[str]):
        """Patches _version_from_hash to provide additional context."""
        _version_from_hash = self._version_from_hash

        def _patch_version_from_hash(*args, **kwargs):
            return _version_from_hash(*args, *additional_context, **kwargs)

        self._version_from_hash = _patch_version_from_hash
        yield
        self._version_from_hash = _version_from_hash

    @property
    def images_client(self) -> image_service_pb2_grpc.ImageServiceStub:
        if self._images_client:
            return self._images_client

        self._images_client = image_service_pb2_grpc.ImageServiceStub(self.client._channel)
        return self._images_client

    @property
    def secret_client(self) -> SecretServiceStub:
        if self._secret_client:
            return self._secret_client

        self._secret_client = SecretServiceStub(self.client._channel)
        return self._secret_client

    def _get_secrets(self, project: Optional[str] = None, domain: Optional[str] = None) -> List[Secret]:
        next_token, has_next = "", True
        org = _get_organization(self.config.platform, channel=self.client._channel)

        secrets = []

        # There are tenants that do not org scoped secrets, so error we assume there are no
        # secrets in that scope
        with suppress(FlyteInvalidInputException):
            while has_next:
                request = ListSecretsRequest(
                    limit=100,
                    organization=org,
                    token=next_token,
                    project=project,
                    domain=domain,
                )
                response: ListSecretsResponse = self.secret_client.ListSecrets(request)
                next_token = response.token
                has_next = next_token != ""

                secrets.extend(response.secrets)

        return secrets

    def _secret_exists(self, secrets: List[Secret], name: str, project: str, domain: str) -> bool:
        """
        Return True if secret name exists that accepts project and domain.
        """
        for secret in secrets:
            if secret.id.name == name:
                if secret.id.project == project and secret.id.domain == domain:
                    return True
                elif secret.id.project == project and secret.id.domain == "":
                    return True
                elif secret.id.project == "" and secret.id.domain == domain:
                    return True
                elif secret.id.project == "" and secret.id.domain == "":
                    return True
        return False

    def _get_image_fqin(self, name: str) -> str:
        if self.config.platform.endpoint.startswith("localhost"):
            # For sandbox testing, assume that we always want to rebuild
            raise FlyteEntityNotExistException("Running in sandbox")
        image_id = image_definition__pb2.ImageIdentifier(name=name)
        org = _get_organization(self.config.platform, channel=self.client._channel)
        req = image_payload__pb2.GetImageRequest(id=image_id, organization=org)
        resp = self.images_client.GetImage(req)
        return resp.image.fqin

    def _get_hf_hub_download_image(self) -> ImageSpec:
        # # TODO: The backend should provide the hfhub-cache image
        return _hf_download_image

    def _get_deployable_engine_container(self, engine) -> str:
        # TODO The backend should provide the serving image
        supported_engines = {"vllm": "ghcr.io/unionai-oss/serving-vllm:0.1.16"}
        if engine not in supported_engines:
            concat_supported_engines = ", ".join(supported_engines)
            msg = f"Supported engines are: {concat_supported_engines}. Got: {engine}"
            raise ValueError(msg)

        return supported_engines[engine]

    def _get_deployable_model(self, uri: str) -> DeployableModel:
        """
        Get a deployable model artifact from Union. This will return the model artifact object.

        :param uri: The URI of the model artifact.
        :return: The DeployableModel object, which contains AppTemplates and the model artifact.
        """
        artifact = self.get_artifact(uri=uri)
        if artifact is None:
            raise ValueError(f"Model artifact not found at URI: {uri}")
        app_templates = get_app_templates_for_model(model=artifact)
        return DeployableModel(artifact=artifact, app_templates=app_templates)

    def _user_info(self) -> UserInfoResponse:
        """
        Query the user info.
        """
        client = IdentityServiceStub(self.client._channel)
        return client.UserInfo(UserInfoRequest())



from union.internal.common import identifier_pb2 as _identifier_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Action(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    ACTION_NONE: _ClassVar[Action]
    ACTION_CREATE: _ClassVar[Action]
    ACTION_READ: _ClassVar[Action]
    ACTION_UPDATE: _ClassVar[Action]
    ACTION_DELETE: _ClassVar[Action]
    ACTION_VIEW_FLYTE_INVENTORY: _ClassVar[Action]
    ACTION_VIEW_FLYTE_EXECUTIONS: _ClassVar[Action]
    ACTION_REGISTER_FLYTE_INVENTORY: _ClassVar[Action]
    ACTION_CREATE_FLYTE_EXECUTIONS: _ClassVar[Action]
    ACTION_ADMINISTER_PROJECT: _ClassVar[Action]
    ACTION_MANAGE_PERMISSIONS: _ClassVar[Action]
    ACTION_ADMINISTER_ACCOUNT: _ClassVar[Action]
    ACTION_MANAGE_CLUSTER: _ClassVar[Action]
    ACTION_EDIT_EXECUTION_RELATED_ATTRIBUTES: _ClassVar[Action]
    ACTION_EDIT_CLUSTER_RELATED_ATTRIBUTES: _ClassVar[Action]
    ACTION_EDIT_UNUSED_ATTRIBUTES: _ClassVar[Action]
ACTION_NONE: Action
ACTION_CREATE: Action
ACTION_READ: Action
ACTION_UPDATE: Action
ACTION_DELETE: Action
ACTION_VIEW_FLYTE_INVENTORY: Action
ACTION_VIEW_FLYTE_EXECUTIONS: Action
ACTION_REGISTER_FLYTE_INVENTORY: Action
ACTION_CREATE_FLYTE_EXECUTIONS: Action
ACTION_ADMINISTER_PROJECT: Action
ACTION_MANAGE_PERMISSIONS: Action
ACTION_ADMINISTER_ACCOUNT: Action
ACTION_MANAGE_CLUSTER: Action
ACTION_EDIT_EXECUTION_RELATED_ATTRIBUTES: Action
ACTION_EDIT_CLUSTER_RELATED_ATTRIBUTES: Action
ACTION_EDIT_UNUSED_ATTRIBUTES: Action

class Organization(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class Domain(_message.Message):
    __slots__ = ["name", "organization"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    name: str
    organization: Organization
    def __init__(self, name: _Optional[str] = ..., organization: _Optional[_Union[Organization, _Mapping]] = ...) -> None: ...

class Project(_message.Message):
    __slots__ = ["name", "domain"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    name: str
    domain: Domain
    def __init__(self, name: _Optional[str] = ..., domain: _Optional[_Union[Domain, _Mapping]] = ...) -> None: ...

class Workflow(_message.Message):
    __slots__ = ["name", "project"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    name: str
    project: Project
    def __init__(self, name: _Optional[str] = ..., project: _Optional[_Union[Project, _Mapping]] = ...) -> None: ...

class LaunchPlan(_message.Message):
    __slots__ = ["name", "project"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    name: str
    project: Project
    def __init__(self, name: _Optional[str] = ..., project: _Optional[_Union[Project, _Mapping]] = ...) -> None: ...

class Resource(_message.Message):
    __slots__ = ["organization", "domain", "project", "workflow", "launch_plan", "cluster"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_FIELD_NUMBER: _ClassVar[int]
    LAUNCH_PLAN_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    organization: Organization
    domain: Domain
    project: Project
    workflow: Workflow
    launch_plan: LaunchPlan
    cluster: _identifier_pb2.ClusterIdentifier
    def __init__(self, organization: _Optional[_Union[Organization, _Mapping]] = ..., domain: _Optional[_Union[Domain, _Mapping]] = ..., project: _Optional[_Union[Project, _Mapping]] = ..., workflow: _Optional[_Union[Workflow, _Mapping]] = ..., launch_plan: _Optional[_Union[LaunchPlan, _Mapping]] = ..., cluster: _Optional[_Union[_identifier_pb2.ClusterIdentifier, _Mapping]] = ...) -> None: ...

class Permission(_message.Message):
    __slots__ = ["resource", "actions"]
    RESOURCE_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    resource: Resource
    actions: _containers.RepeatedScalarFieldContainer[Action]
    def __init__(self, resource: _Optional[_Union[Resource, _Mapping]] = ..., actions: _Optional[_Iterable[_Union[Action, str]]] = ...) -> None: ...



from typing import Optional

import rich_click as click
from grpc import RpcError

from union._config import _get_config_obj
from union.app import App
from union.cli._common import _get_channel_with_org
from union.cli._option import MutuallyExclusiveArgument, MutuallyExclusiveOption
from union.internal.secret.definition_pb2 import SecretIdentifier, SecretSpec
from union.internal.secret.payload_pb2 import UpdateSecretRequest
from union.internal.secret.secret_pb2_grpc import SecretServiceStub


@click.group(name="update")
def update():
    """Update a resource."""


@update.command()
@click.argument(
    "name",
    required=False,
    cls=MutuallyExclusiveArgument,
    mutually_exclusive=["name_option"],
    error_msg="Please pass --name once: `union update secret --name NAME`",
)
@click.option(
    "--name",
    "name_option",
    help="Secret name",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["name"],
    error_msg="Please pass --name once: `union update secret --name NAME`",
)
@click.option(
    "--value",
    help="Secret value",
    prompt="Enter secret value",
    hide_input=True,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value_file"],
)
@click.option(
    "-f",
    "--value-file",
    help="Path to file containing the secret",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=True),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value"],
)
@click.option("--project", help="Project name")
@click.option("--domain", help="Domain name")
def secret(
    name: Optional[str],
    name_option: Optional[str],
    value: str,
    value_file: str,
    project: Optional[str],
    domain: Optional[str],
):
    """Update secret with NAME."""
    name = name or name_option

    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    if value_file:
        with open(value_file, "rb") as f:
            secret_spec = SecretSpec(binary_value=f.read())
    else:
        secret_spec = SecretSpec(string_value=value)

    stub = SecretServiceStub(channel)
    request = UpdateSecretRequest(
        id=SecretIdentifier(name=name, domain=domain, project=project, organization=org),
        secret_spec=secret_spec,
    )

    try:
        stub.UpdateSecret(request)
        click.echo(f"Updated secret with name: {name}")
    except RpcError as e:
        raise click.ClickException(f"Unable to update secret with name: {name}\n{e}") from e


def update_application(app: App, project: str, domain: str):
    from union.configuration import UnionAIPlugin

    remote = UnionAIPlugin.get_remote(config=None, project=project, domain=domain)
    remote._app_remote.update(app)



import base64
import os
import sys
import textwrap
from dataclasses import dataclass
from enum import Enum
from functools import partial
from importlib.metadata import version
from os import getenv
from pathlib import Path
from string import Template
from textwrap import dedent
from typing import List, Optional, Tuple, Union

import click
import grpc
from flyteidl.service.identity_pb2 import UserInfoRequest
from flyteidl.service.identity_pb2_grpc import IdentityServiceStub
from flytekit.clients.auth_helper import get_authenticated_channel
from flytekit.configuration import AuthType, Config, PlatformConfig, get_config_file
from flytekit.configuration.default_images import DefaultImages

_DEFAULT_GCP_SERVERLESS_ENDPOINT: str = "serverless-1.us-east-2.s.union.ai"
_VANITY_UNION_URLS: dict[str, str] = {
    "serverless.union.ai": _DEFAULT_GCP_SERVERLESS_ENDPOINT,
    "serverless.staging.union.ai": "serverless-gcp.cloud-staging.union.ai",
    "serverless.canary.union.ai": "serverless-preview.canary.unionai.cloud",
}
_SERVERLESS_ENDPOINTS = {
    "serverless-1.us-east-2.s.union.ai",
    "serverless-gcp.cloud-staging.union.ai",
    "utt-srv-staging-1.cloud-staging.union.ai",
    "serverless-preview.canary.unionai.cloud",
    "utt-srv-canary-1.canary.unionai.cloud",
}

_UNION_CONFIG_ENV_VARS: List[str] = ["UNION_CONFIG", "UNIONAI_CONFIG"]
_UNION_SERVERLESS_ENDPOINT_ENV_VAR: str = "UNION_SERVERLESS_ENDPOINT"
_UNION_EAGER_API_KEY_ENV_VAR: str = "_UNION_EAGER_API_KEY"
_UNION_API_KEY_ENV_VARS: List[str] = [
    "UNION_API_KEY",
    "UNION_SERVERLESS_API_KEY",
    "UNIONAI_SERVERLESS_API_KEY",
    _UNION_EAGER_API_KEY_ENV_VAR,
    "_UNION_UNION_API_KEY",
]
_UNION_DEFAULT_CONFIG_PATHS: List[Path] = [
    Path.home() / ".union" / "config.yaml",
    Path.home() / ".uctl" / "config.yaml",
]
_UNION_DEFAULT_CONFIG_PATH: Path = _UNION_DEFAULT_CONFIG_PATHS[0]
_DEFAULT_PROJECT_BYOC: str = "flytesnacks"
_DEFAULT_PROJECT_SERVERLESS: str = "default"
_DEFAULT_DOMAIN: str = "development"


@dataclass
class _UnionConfig:
    org: Optional[str] = None
    is_direct_union_cli_call: bool = False

    # These fields are populated when _get_config_obj is called.
    config_obj: Optional[Config] = None
    config_source: str = ""


_UNION_CONFIG = _UnionConfig()


def _is_serverless_endpoint(endpoint: str) -> bool:
    """Check if endpoint is serverless."""
    return endpoint in (endpoint, f"dns:///{endpoint}") and (
        endpoint in _SERVERLESS_ENDPOINTS or getenv(_UNION_SERVERLESS_ENDPOINT_ENV_VAR)
    )


def _get_image_builder_priority(endpoint: str) -> int:
    """Return image builder priority for union image builder.

    - Serverless gets the highest priority (10)
    - BYOC gets a priority (0) lower than the default OSS image builder (1). This
      keeps backward compatibility.
    """
    return 10 if _is_serverless_endpoint(endpoint) else 0


@dataclass
class AppClientCredentials:
    endpoint: str
    client_id: str
    client_secret: str
    org: str


def _encode_app_client_credentials(app_credentials: AppClientCredentials) -> str:
    """Encode app_credentials with base64."""
    data = (
        f"{app_credentials.endpoint}:{app_credentials.client_id}:{app_credentials.client_secret}:{app_credentials.org}"
    )
    return base64.b64encode(data.encode("utf-8")).decode("utf-8")


def _decode_app_client_credentials(encoded_str: str) -> AppClientCredentials:
    """Decode encoded base64 string into app credentials."""
    endpoint, client_id, client_secret, org = base64.b64decode(encoded_str.encode("utf-8")).decode("utf-8").split(":")
    # In byoc, the org is encoded as the 'None' string which we want to marshal back to a proper None type.
    if org == "None" or not _is_serverless_endpoint(endpoint):
        org = None
    return AppClientCredentials(endpoint=endpoint, client_id=client_id, client_secret=client_secret, org=org)


def _clean_endpoint(endpoint: str) -> str:
    """Clean endpoint."""
    prefixes = ("dns:///", "http://", "https://")

    for prefix in prefixes:
        if endpoint.startswith(prefix):
            n_prefix = len(prefix)
            endpoint = endpoint[n_prefix:]
            break

    return endpoint.rstrip("/")


def _get_endpoint_for_login(*, host: Optional[str] = None, serverless: bool = False) -> str:
    """Get endpoint for login."""
    if host is None and not serverless:
        msg = "--host HOST or --serverless must be passed in for login"
        raise click.ClickException(msg)

    if host is not None and serverless:
        msg = "Can not pass in both --host and --serverless for login, please choose one"
        raise click.ClickException(msg)

    if serverless:
        host = getenv(_UNION_SERVERLESS_ENDPOINT_ENV_VAR, default=_DEFAULT_GCP_SERVERLESS_ENDPOINT)

    endpoint = _clean_endpoint(host)

    if endpoint in _VANITY_UNION_URLS:
        return _VANITY_UNION_URLS[endpoint]
    return endpoint


def _get_env_var(env_vars: List[str]) -> Tuple[Optional[str], Optional[str]]:
    """Get environment variable from a list of env_vars."""
    for env_var in env_vars:
        if env_var in os.environ:
            return os.environ[env_var], env_var
    return None, None


def _get_config_path(config_paths: List[Path] = _UNION_DEFAULT_CONFIG_PATHS) -> Optional[Path]:
    """Get configuration path."""
    for path in config_paths:
        if path.exists():
            return path
    return None


_get_union_config_env_var = partial(_get_env_var, env_vars=_UNION_CONFIG_ENV_VARS)
_get_union_api_env_var = partial(_get_env_var, env_vars=_UNION_API_KEY_ENV_VARS)


def _check_yaml_config_is_empty(config_file: Union[str, os.PathLike]):
    with open(config_file, "r") as f:
        contents = f.read().strip()

    if not contents:
        raise click.ClickException(
            f"Unable to load yaml at {config_file} because it is empty. Delete file or pass a config that is not empty"
        )


def _decode_union_api_key(serverless_api_value: str) -> Config:
    """Decode union serverless api key and return Config."""
    config = Config.auto()
    try:
        app_credentials = _decode_app_client_credentials(serverless_api_value)
    except Exception as e:
        raise ValueError("Unable to read UNION_API_KEY") from e

    if _UNION_CONFIG.org is None and app_credentials.org != "":
        _UNION_CONFIG.org = app_credentials.org

    return config.with_params(
        platform=PlatformConfig(
            endpoint=app_credentials.endpoint,
            insecure=False,
            auth_mode=AuthType.CLIENTSECRET,
            client_id=app_credentials.client_id,
            client_credentials_secret=app_credentials.client_secret,
        )
    )


def _should_default_to_union_semantics() -> bool:
    return bool(_UNION_CONFIG.is_direct_union_cli_call or os.getenv(_UNION_EAGER_API_KEY_ENV_VAR))


class ConfigSource(Enum):
    CLI = "--config passed in"
    REMOTE = "Set by Remote"


@dataclass
class ConfigWithSource:
    config: Union[Config, str]
    source: ConfigSource


def _get_config_obj(
    config_in: Optional[Union[str, ConfigWithSource]] = None, default_to_union_semantics: bool = False
) -> Config:
    """Get Config object.

    If `config_in` is not None, then it will be used as the Config file.

    If the `union` CLI is called directly or `default_to_union_semantics=True`, then the config_in is set
    in the following order:
        1. `UNION_SERVERLESS_ENDPOINT` environment variable (Used for development)
        2. `UNION_SERVERLESS_API_KEY` environment variable
        3. `UNION_CONFIG` environment variable
        4. ~/.union/config.yaml if it exists

    If `pyflyte` CLI is called and `flytekit`'s `get_config_file` did not return a `ConfigFile`,
    then serverless it the default endpoint.
    """
    default_to_union_semantics = default_to_union_semantics or _should_default_to_union_semantics()

    config_out = None
    if config_in is not None:
        if isinstance(config_in, str):
            config_out = Config.auto(config_in)
        else:  # ConfigWithSource object
            if isinstance(config_in.config, str):
                config_out = Config.auto(config_in.config)
            else:  # Config object
                config_out = config_in.config
            _UNION_CONFIG.config_source = config_in.source.value

    elif _UNION_CONFIG.config_obj is not None and _UNION_CONFIG.config_source == ConfigSource.CLI.value:
        # Config came from --config, it will have the highest precedence.
        config_out = _UNION_CONFIG.config_obj

    elif default_to_union_semantics:
        if (endpoint := getenv(_UNION_SERVERLESS_ENDPOINT_ENV_VAR)) is not None:
            # If UNION_SERVERLESS_ENDPOINT is set in `_UnionConfig`, use it instead
            _UNION_CONFIG.config_source = f"{_UNION_SERVERLESS_ENDPOINT_ENV_VAR} env variable"
            config_out = Config.for_endpoint(endpoint=endpoint)

        elif (api_value_tuple := _get_union_api_env_var()) and (api_value_tuple[0] not in ("", None)):
            _UNION_CONFIG.config_source = f"{api_value_tuple[1]} env variable"
            config_out = _decode_union_api_key(api_value_tuple[0])

        elif (api_config_tuple := _get_union_config_env_var()) and (api_config_tuple[0] not in ("", None)):
            _UNION_CONFIG.config_source = f"{api_config_tuple[1]} env variable"
            config_out = Config.auto(api_config_tuple[0])

        elif (config_path := _get_config_path()) is not None:
            # Check if config is empty:
            _check_yaml_config_is_empty(config_path)
            # Read config from file system
            _UNION_CONFIG.config_source = f"{config_path} file"
            config_out = Config.auto(str(config_path.absolute()))

    elif _UNION_CONFIG.config_obj is not None and _UNION_CONFIG.config_source == ConfigSource.REMOTE.value:
        # Configured through remote, use this configuration if none of the automatic configurations are found
        config_out = _UNION_CONFIG.config_obj

    if config_out is None:
        # Raise union config error if directly called by CLI or default_to_union_semantics
        is_local_cli = os.getenv("FLYTE_INTERNAL_EXECUTION_ID") is None

        if is_local_cli and default_to_union_semantics:
            msg = dedent("""\
            Please login to Union.

            - On Union Serverless run: `union create login --serverless`
            - On Union BYOC run: `union create login --host UNION_TENANT`

            You can also pass a configuration with `union --config` or set UNION_CONFIG.\
            """)
            from union._exceptions import UnionRequireConfigException

            raise UnionRequireConfigException(msg)

        # Otherwise, try flyte configuration
        cfg_file = get_config_file(config_in)

        if cfg_file is None:
            if is_local_cli:
                from union._exceptions import UnionRequireConfigException

                msg = dedent("""\
                Please configure pyflyte with FLYTECTL_CONFIG or pass in --config.

                Alternatively, use the union CLI and login with `union create login`:

                - On Union Serverless run: `union create login --serverless`
                - On Union BYOC run: `union create login --host UNION_TENANT`\
                """)
                raise UnionRequireConfigException(msg)
            config_out = Config.for_sandbox()
        else:
            config_out = Config.auto(cfg_file)

    _UNION_CONFIG.config_obj = config_out
    return config_out


def _config_from_api_key(api_key: str) -> Config:
    """Get config from api key."""
    config_out = _decode_union_api_key(api_key)

    _UNION_CONFIG.config_source = "manual API key"
    _UNION_CONFIG.config_obj = config_out
    return config_out


def _get_organization(platform_config: PlatformConfig, channel: Optional[grpc.Channel] = None) -> Optional[str]:
    """Get organization based on endpoint."""
    if _UNION_CONFIG.org is not None:
        return _UNION_CONFIG.org
    elif _is_serverless_endpoint(platform_config.endpoint):
        org = _get_user_handle(platform_config, channel)
        _UNION_CONFIG.org = org
        return org
    else:
        # Managed+ users, the org is not required for requests and we set it None
        # to replicate default flytekit behavior.
        return None


def _get_user_handle(platform_config: PlatformConfig, channel: Optional[grpc.Channel] = None) -> str:
    """Get user_handle for PlatformConfig."""
    if channel is None:
        channel = get_authenticated_channel(platform_config)

    client = IdentityServiceStub(channel)
    user_info = client.UserInfo(UserInfoRequest())
    user_handle = user_info.additional_claims.fields["userhandle"]
    return user_handle.string_value


def _get_default_image() -> str:
    """Get default image version."""
    from union._exceptions import UnionRequireConfigException

    try:
        cfg_obj = _get_config_obj(default_to_union_semantics=False)

        # TODO: This is only temporary to support GCP endpoints. When the union images are public,
        # we will always use union images
        endpoint = cfg_obj.platform.endpoint
        if _is_serverless_endpoint(endpoint):
            major, minor = sys.version_info.major, sys.version_info.minor
            union_version = version("union")
            if "dev" in union_version:
                suffix = "latest"
            else:
                suffix = union_version

            return f"cr.union.ai/v1/unionai/union:py{major}.{minor}-{suffix}"
        else:
            return DefaultImages().find_image_for()

    except UnionRequireConfigException:
        return DefaultImages().find_image_for()


def _write_config_to_path(endpoint: str, auth_type: str, config_path: Path = _UNION_DEFAULT_CONFIG_PATH):
    """Write config to config directory."""
    config_dir = config_path.parent
    config_dir.mkdir(exist_ok=True, parents=True)

    config_template = Template(
        textwrap.dedent(
            """\
    admin:
      endpoint: $endpoint
      insecure: false
      authType: $auth_type
    logger:
      show-source: true
      level: 0
    union:
      connection:
        host: $endpoint
        insecure: false
      auth:
        type: $auth_type
    """
        )
    )
    config_path.write_text(config_template.substitute(endpoint=f"dns:///{endpoint}", auth_type=auth_type))
    return config_path


def _get_default_project(previous_default: str, cfg_obj: Optional[Config] = None) -> str:
    from union._exceptions import UnionRequireConfigException

    try:
        if cfg_obj is None:
            cfg_obj = _get_config_obj()
        if _is_serverless_endpoint(cfg_obj.platform.endpoint):
            return _DEFAULT_PROJECT_SERVERLESS
    except UnionRequireConfigException:
        return previous_default

    return previous_default


def _get_auth_success_html(endpoint: str) -> str:
    """Get default success html. Return None to use flytekit's default success html."""
    if endpoint.endswith("union.ai") or endpoint.endswith("unionai.cloud"):
        SUCCESS_HTML = textwrap.dedent(
            f"""
        <html>
        <head>
            <title>OAuth2 Authentication to Union Successful</title>
        </head>
        <body style="background:white;font-family:Arial">
            <div style="position: absolute;top:40%;left:50%;transform: translate(-50%, -50%);text-align:center;">
                <div style="margin:auto">
                    <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 300 65" fill="currentColor"
                        style="color:#fdb51e;width:360px;">
                        <title>Union.ai</title>
                        <path d="M32,64.8C14.4,64.8,0,51.5,0,34V3.6h17.6v41.3c0,1.9,1.1,3,3,3h23c1.9,0,3-1.1,3-3V3.6H64V34
                        C64,51.5,49.6,64.8,32,64.8z M69.9,30.9v30.4h17.6V20c0-1.9,1.1-3,3-3h23c1.9,0,3,1.1,3,3v41.3H134V30.9c0-17.5-14.4-30.8-32.1-30.8
                        S69.9,13.5,69.9,30.9z M236,30.9v30.4h17.6V20c0-1.9,1.1-3,3-3h23c1.9,0,3,1.1,3,3v41.3H300V30.9c0-17.5-14.4-30.8-32-30.8
                        S236,13.5,236,30.9L236,30.9z M230.1,32.4c0,18.2-14.2,32.5-32.2,32.5s-32-14.3-32-32.5s14-32.1,32-32.1S230.1,14.3,230.1,32.4
                        L230.1,32.4z M213.5,20.2c0-1.9-1.1-3-3-3h-24.8c-1.9,0-3,1.1-3,3v24.5c0,1.9,1.1,3,3,3h24.8c1.9,0,3-1.1,3-3V20.2z M158.9,3.6
                        h-17.6v57.8h17.6V3.6z"></path>
                    </svg>
                    <h2>You've successfully authenticated to Union!</h2>
                    <p style="font-size:20px;">Return to your terminal for next steps</p>
                </div>
            </div>
        </body>
        </html>
        """  # noqa
        )
        return SUCCESS_HTML
    return None



from union.internal.common import authorization_pb2 as _authorization_pb2
from union.internal.common import identifier_pb2 as _identifier_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Policy(_message.Message):
    __slots__ = ["id", "bindings", "description"]
    ID_FIELD_NUMBER: _ClassVar[int]
    BINDINGS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    id: _identifier_pb2.PolicyIdentifier
    bindings: _containers.RepeatedCompositeFieldContainer[PolicyBinding]
    description: str
    def __init__(self, id: _Optional[_Union[_identifier_pb2.PolicyIdentifier, _Mapping]] = ..., bindings: _Optional[_Iterable[_Union[PolicyBinding, _Mapping]]] = ..., description: _Optional[str] = ...) -> None: ...

class PolicyBinding(_message.Message):
    __slots__ = ["role_id", "resource"]
    ROLE_ID_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_FIELD_NUMBER: _ClassVar[int]
    role_id: _identifier_pb2.RoleIdentifier
    resource: _authorization_pb2.Resource
    def __init__(self, role_id: _Optional[_Union[_identifier_pb2.RoleIdentifier, _Mapping]] = ..., resource: _Optional[_Union[_authorization_pb2.Resource, _Mapping]] = ...) -> None: ...



from typing import Optional

import rich_click as click
from grpc import RpcError

from union._config import _DEFAULT_DOMAIN, _DEFAULT_PROJECT_BYOC, _UNION_DEFAULT_CONFIG_PATH, _get_config_obj
from union.cli._common import _get_channel_with_org
from union.cli._option import MutuallyExclusiveArgument, MutuallyExclusiveOption
from union.internal.identity.app_payload_pb2 import DeleteAppRequest, GetAppRequest
from union.internal.identity.app_service_pb2_grpc import AppsServiceStub
from union.internal.secret.definition_pb2 import SecretIdentifier
from union.internal.secret.payload_pb2 import DeleteSecretRequest
from union.internal.secret.secret_pb2_grpc import SecretServiceStub
from union.remote import UnionRemote
from union.workspace._vscode_remote import _get_workspace_executions


@click.group()
def delete():
    """Delete a resource."""


@delete.command()
@click.argument(
    "name",
    required=False,
    cls=MutuallyExclusiveArgument,
    mutually_exclusive=["name_option"],
    error_msg="Please pass --name once: `union delete secret --name NAME`",
)
@click.option(
    "--name",
    "name_option",
    help="Secret name",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["name"],
    error_msg="Please pass --name once: `union create secret --name NAME`",
)
@click.option("--project", help="Project name")
@click.option("--domain", help="Domain name")
def secret(
    name: Optional[str],
    name_option: Optional[str],
    project: Optional[str],
    domain: Optional[str],
):
    """Delete secret with NAME."""
    name = name or name_option

    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    stub = SecretServiceStub(channel)

    request = DeleteSecretRequest(
        id=SecretIdentifier(name=name, domain=domain, project=project, organization=org),
    )
    try:
        stub.DeleteSecret(request)
        click.echo(f"Deleted secret with name: {name}")
    except RpcError as e:
        raise click.ClickException(f"Unable to delete secret with name: {name}\n{e}") from e


@delete.group("api-key")
def api_key():
    """Manage API keys."""


@api_key.command("admin")
@click.option("--name", type=str, help="Name for API key", required=True)
def admin(name: str):
    """Delete api key."""
    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    stub = AppsServiceStub(channel)

    # Get app first to make sure it exist
    get_request = GetAppRequest(organization=org, client_id=name)
    try:
        stub.Get(get_request)
    except RpcError as e:
        raise click.ClickException(f"Unable to delete api-key with name: {name} because it does not exist.\n{e}") from e

    delete_request = DeleteAppRequest(organization=org, client_id=name)
    try:
        stub.Delete(delete_request)
        click.echo(f"Deleted api-key with name: {name}")
    except RpcError as e:
        raise click.ClickException(f"Unable to delete api-key with name: {name}\n{e}") from e


@delete.command()
def login():
    """Delete login information."""
    if not _UNION_DEFAULT_CONFIG_PATH.exists():
        click.echo(f"Login at {_UNION_DEFAULT_CONFIG_PATH} does not exist. No need to delete.")
        return

    try:
        _UNION_DEFAULT_CONFIG_PATH.unlink()
        click.echo("Deleted login.")
    except Exception as e:
        msg = f"Unable to delete login.\n{e}"
        raise click.ClickException(msg) from e


@delete.command()
@click.pass_context
@click.argument("name")
@click.option("--project", default=_DEFAULT_PROJECT_BYOC, help="Project name")
@click.option("--domain", default=_DEFAULT_DOMAIN, help="Domain name")
def workspace(ctx: click.Context, name: str, project: str, domain: str):
    """Delete workspace with NAME."""
    remote = UnionRemote(default_domain=domain, default_project=project)

    executions = _get_workspace_executions(remote, project, domain)

    for execution, workspace_name in executions:
        if workspace_name == name:
            remote.terminate(execution, cause="Stopped from CLI")
            click.echo(f"Workspace {name} deleted!")
            break
    else:  # no break
        click.echo(f"Workspace with name: {name} does not exist or no longer running")



from union.internal.app import app_definition_pb2 as _app_definition_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ReplicaIdentifier(_message.Message):
    __slots__ = ["app_id", "name"]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    app_id: _app_definition_pb2.Identifier
    name: str
    def __init__(self, app_id: _Optional[_Union[_app_definition_pb2.Identifier, _Mapping]] = ..., name: _Optional[str] = ...) -> None: ...

class ReplicaMeta(_message.Message):
    __slots__ = ["id", "revision"]
    ID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    id: ReplicaIdentifier
    revision: int
    def __init__(self, id: _Optional[_Union[ReplicaIdentifier, _Mapping]] = ..., revision: _Optional[int] = ...) -> None: ...

class ReplicaList(_message.Message):
    __slots__ = ["items"]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[Replica]
    def __init__(self, items: _Optional[_Iterable[_Union[Replica, _Mapping]]] = ...) -> None: ...

class Replica(_message.Message):
    __slots__ = ["metadata", "status"]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    metadata: ReplicaMeta
    status: ReplicaStatus
    def __init__(self, metadata: _Optional[_Union[ReplicaMeta, _Mapping]] = ..., status: _Optional[_Union[ReplicaStatus, _Mapping]] = ...) -> None: ...

class ReplicaStatus(_message.Message):
    __slots__ = ["deployment_status", "reason"]
    DEPLOYMENT_STATUS_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    deployment_status: str
    reason: str
    def __init__(self, deployment_status: _Optional[str] = ..., reason: _Optional[str] = ...) -> None: ...



import datetime
import typing
from typing import Union

from flyteidl.core import artifact_id_pb2 as art_id
from flytekit.core.artifact import ArtifactQuery, Partitions, TimePartition
from google.protobuf.timestamp_pb2 import Timestamp

from union.internal.artifacts import artifacts_pb2


def _construct_artifact_key(
    project: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    name: typing.Optional[str] = None,
    artifact_key: typing.Optional[art_id.ArtifactKey] = None,
    query: typing.Optional[ArtifactQuery] = None,
) -> art_id.ArtifactKey:
    if artifact_key:
        ak = artifact_key
    elif query:
        p = query.artifact.project or project
        d = query.artifact.domain or domain
        n = query.artifact.name  # if a query is given, name should be specified in the artifact
        ak = _get_artifact_key(p, d, n)
    else:
        ak = _get_artifact_key(project, domain, name)
    return ak


def _construct_partitions(
    query: typing.Optional[ArtifactQuery] = None,
    partitions: typing.Optional[Union[Partitions, typing.Dict[str, str]]] = None,
) -> typing.Optional[art_id.Partitions]:
    idl_partitions = None
    if query and query.partitions:
        if partitions:
            raise ValueError("Cannot specify partitions in both query and as an argument.")
        idl_partitions = query.partitions.to_flyte_idl()
    elif isinstance(partitions, Partitions):
        idl_partitions = partitions.to_flyte_idl()
    elif partitions:
        idl_partitions = Partitions(partitions).to_flyte_idl()
    return idl_partitions


def _construct_time_stamp(
    query: typing.Optional[ArtifactQuery] = None,
    time_partition: typing.Optional[Union[datetime.datetime, TimePartition]] = None,
) -> typing.Optional[Timestamp]:
    idl_time_partition = None
    if query and query.time_partition:
        if time_partition:
            raise ValueError("Cannot specify time partition in both query and as an argument.")
        idl_time_partition = query.time_partition.value.time_value
    elif isinstance(time_partition, TimePartition):
        idl_time_partition = time_partition.to_flyte_idl().value.time_value
    elif time_partition:
        if not isinstance(time_partition, datetime.datetime):
            raise ValueError("Time partition must be a datetime or TimePartition object.")
        idl_time_partition = Timestamp()
        idl_time_partition.FromDatetime(time_partition)
    return idl_time_partition


def construct_search_artifact_request(
    project: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    name: typing.Optional[str] = None,
    artifact_key: typing.Optional[art_id.ArtifactKey] = None,
    query: typing.Optional[ArtifactQuery] = None,
    partitions: typing.Optional[Union[Partitions, typing.Dict[str, str]]] = None,
    time_partition: typing.Optional[Union[datetime.datetime, TimePartition]] = None,
    group_by_key: bool = False,
    limit: int = 100,
    token: typing.Optional[str] = None,
) -> artifacts_pb2.SearchArtifactsRequest:
    # Construct key - must have either an artifact key or project, domain, name specified, or a query input
    ak = _construct_artifact_key(project, domain, name, artifact_key, query)

    # Construct partitions object, if applicable, can be None
    idl_partitions = _construct_partitions(query, partitions)

    # Construct time partition object, if applicable, can be None
    idl_time_partition = _construct_time_stamp(query, time_partition)

    # Construct search options based on input args
    search_options = artifacts_pb2.SearchOptions(
        strict_partitions=False,  # always set this to false so the user doesn't have to specify all partitions
        latest_by_key=group_by_key,
    )

    search_request = artifacts_pb2.SearchArtifactsRequest(
        artifact_key=ak,
        partitions=idl_partitions,
        time_partition_value=idl_time_partition,
        options=search_options,
        limit=limit,
        token=token,
        version=None,  # leave this empty, if you have the version, use get instead.
        principal=None,  # leave this empty for now, we don't have a nice way to get okta id
        granularity=None,  # leave this empty for now until time partition is handled
    )
    return search_request


def _get_artifact_key(
    project: typing.Optional[str] = None,
    domain: typing.Optional[str] = None,
    name: typing.Optional[str] = None,
) -> art_id.ArtifactKey:
    if not project or not domain or not name:
        raise ValueError("Project, domain, and name are required to create an artifact key")

    return art_id.ArtifactKey(
        project=project,
        domain=domain,
        name=name,
    )
